package org.playonline.network.client;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import org.playonline.network.client.custom.SecondUGASPClient;
import org.playonline.ui.basic.map.handler.MapViewMessage;
import org.playonline.ui.basic.profile.Character;

import com.google.android.maps.GeoPoint;

import android.os.Message;
import android.util.Log;

/**
 * This class holds the methods that will be invoked remotely by other clients 
 * or by game logic.
 * 
 * Declare methods respecting following properties:
 * <ul>
 * <li>they must be instances methods, and not static ones</li>
 * <li>types handled as parameters are Byte, Short, Integer, Long, Float, Double, Boolean, String, RMAISerializable and arrays of all those types</li>
 * <li>as method invocations through RMAI protocol are asynchronous,
 * only void methods can be invoked. However a callback mechanism is provided, allowing to invoke remote method
 * on the entity which did the invocation first (as it is done in replyToMessage(String message))</li>
 * </ul>
 * 
 * Entity which has invoked a method located in this class can be retrieved, using getSenderActorSessionID().
 * 
 * @see replyToMessage(String message)
 **/
public class RMAIClientManager {

	private UGASPClient client;
	private short senderActorSessionID;
	private short recipientActorSessionID;

	public RMAIClientManager(UGASPClient client, short senderActorSessionID, short recipientActorSessionID){
		this.client = client;
		this.senderActorSessionID = senderActorSessionID;
		this.recipientActorSessionID = recipientActorSessionID;
	}

	public UGASPClient getClient() {
		return client;
	}

	public short getSenderActorSessionID() {
		return senderActorSessionID;
	}

	public short getRecipientActorSessionID() {
		return recipientActorSessionID;
	}

	/** 
	 * This class is used whether remote procedures should be invoked asynchronously on uGASP game logic, 
	 * or on every uGASP client, or on both.
	 * If you want to do so, in the several GASPClient.remoteMethodAsynchronousInvocation(int recipientActorSessionID,String methodName,Object param0,...),
	 * you should use either:
	 * <ul>
	 * <li>ActorSessionID.uGASP_GAME_LOGIC.getID(), or</li>
	 * <li>ActorSessionID.BROADCAST_WITHOUT_GAME_LOGIC.getID(), or</li> 
	 * <li>ActorSessionID.BROADCAST_WITH_GAME_LOGIC.getID()</li>
	 * </ul>
	 * as the recipientActorSessionID.
	 * If you just want to invoke a method on one client, simply use its actor session ID (the int value). 
	 */
	public static class ActorSessionID{

		private final short id;

		private ActorSessionID(short id){
			this.id = id;
		}

		public short getID(){return id;}
		public final static ActorSessionID uGASP_GAME_LOGIC = new ActorSessionID((short)0);
		public final static ActorSessionID BROADCAST_WITHOUT_GAME_LOGIC = new ActorSessionID((short)-1);
		public final static ActorSessionID BROADCAST_WITH_GAME_LOGIC = new ActorSessionID((short)-2);


		public String toString(){
			// for uGASP game logic
			if(id == uGASP_GAME_LOGIC.getID())return "uGASP_GAME_LOGIC";
			// for broadcast to clients
			else if(id == BROADCAST_WITHOUT_GAME_LOGIC.getID())return "BROADCAST_WITHOUT_GAME_LOGIC";
			// for broadcast to clients and server
			else if(id == BROADCAST_WITH_GAME_LOGIC.getID())return "BROADCAST_WITH_GAME_LOGIC";
			// otherwise 
			else return "INCORRECT_ACTOR_SESSION_ID";
		}


	}


	/*********************************************************************************************************/
	/*************************************    Remote Methods    **********************************************/
	/*********************************************************************************************************/


	public void displayToastMessage(String message){
		Log.d(this.getClass().getSimpleName(), "RMAI called =  displayToastMessage : \n"+message);
		final String finalMessage = message;
		Message msg = MapViewMessage.displayLongToastMessage(finalMessage);
		((SecondUGASPClient)client).getMapActivity().getMapViewHandler().sendMessage(msg);
	}

	
	public void displayAlertMessage(String message){
		Log.d(this.getClass().getSimpleName(), "RMAI called =  displayAlertMessage : \n"+message);
		final String finalMessage = message;
		Message msg = MapViewMessage.displayAlertMessage(finalMessage);
		((SecondUGASPClient)client).getMapActivity().getMapViewHandler().sendMessage(msg);
	}

	
	public void initializeCharactersListOnLocationMap(Object[] locationEntries,Object[] charactersListValues){
		if(locationEntries.length == charactersListValues.length){
			Map<GeoPoint,List<Short>> map = new ConcurrentHashMap<GeoPoint, List<Short>>();
			GeoPoint[] geoPointArray = RMAIArraysUtil.toGeoPointArray(locationEntries);
			for (int i = 0; i < geoPointArray.length; i++) {
				List<Short> charactersList = RMAIArraysUtil.toListShort((Object[])charactersListValues[i]);
				map.put(geoPointArray[i], charactersList);
			}
			((SecondUGASPClient)client).getMapActivity().getCharactersManager().setCharactersListInLocationMap(map);
		}
		
	}
	
	
	public void initializeCharacterByActorSessionIdMap(Object[] actorsSessionIdsEntries,Object[] characterNameValues,Object[] isAttachedToLeaderValues,Object[] leaderActorSessionIdValues){
		if(actorsSessionIdsEntries.length == characterNameValues.length &&  isAttachedToLeaderValues.length == leaderActorSessionIdValues.length && actorsSessionIdsEntries.length == isAttachedToLeaderValues.length){
			Map<Short,Character> map = new ConcurrentHashMap<Short, Character>();
			for (int i = 0; i < actorsSessionIdsEntries.length; i++) {
				Short actorSessionId = (Short)actorsSessionIdsEntries[i];
				String characterName = (String)characterNameValues[i];
				Boolean isAttachedToLeader = (Boolean)isAttachedToLeaderValues[i];
				Short leaderActorSessionId = (Short)leaderActorSessionIdValues[i];
				Character character = Character.getInstance(characterName);
				if(isAttachedToLeader){
					character.attachToLeader(leaderActorSessionId.shortValue());
				}
				map.put(actorSessionId, character);
			}
			// if its own character is present on the map, it is removed
			if(map.containsKey(Short.valueOf(client.getID().getActorSessionID()))){
				map.remove(Short.valueOf(client.getID().getActorSessionID()));
			}
			((SecondUGASPClient)client).getMapActivity().getCharactersManager().setCharacterByActorSessionIdMap(map);
		}
	}
	
	
	public void storeCharacter(Integer latitudeE6, Integer longitudeE6, Short characterActorSessionId, String characterName){
		// if character is not my own character
		if(characterActorSessionId.shortValue() != ((SecondUGASPClient)client).getID().getActorSessionID()){
			((SecondUGASPClient)client).getMapActivity().getCharactersManager().putCharacter(characterActorSessionId, characterName);
			GeoPoint location = new GeoPoint(latitudeE6.intValue(), longitudeE6.intValue());
			((SecondUGASPClient)client).getMapActivity().getCharactersManager().moveCharacterToLocation(location,characterActorSessionId);
		}
	}

	
	public void updateCharacterLocation(Integer latitudeE6, Integer longitudeE6, Short characterActorSessionId){
		// if character is not my own character
		if(characterActorSessionId.shortValue() != ((SecondUGASPClient)client).getID().getActorSessionID()){
			GeoPoint location = new GeoPoint(latitudeE6.intValue(), longitudeE6.intValue());
			((SecondUGASPClient)client).getMapActivity().getCharactersManager().moveCharacterToLocation(location,characterActorSessionId);
		}
	}
	
	
	public void characterJoinLeader(Short memberActorSessionId, Short leaderActorSessionId) {
		// if member is not my own character
		if(client.getID().getActorSessionID() != memberActorSessionId.shortValue()){
			// informations about characters are are stored
			((SecondUGASPClient)client).getMapActivity().getCharactersManager().getCharacterByActorSessionId(memberActorSessionId).attachToLeader(leaderActorSessionId.shortValue());
		}
	}
	
	
	public void characterLeaveLeader(Short memberActorSessionId) {
		// if member is not my own character
		if(client.getID().getActorSessionID() != memberActorSessionId.shortValue()){
			// informations about characters are are stored
			((SecondUGASPClient)client).getMapActivity().getCharactersManager().getCharacterByActorSessionId(memberActorSessionId).detachFromLeader();
		}
	}
	
	/**********************************************************************************************************************/
	/************************************ Dialog management for group members *********************************************/
	/**********************************************************************************************************************/
	
	public void showDialog(Short dialogId){
		Log.d(getClass().getSimpleName(), "--- RMAI called = showDialog "+dialogId);
		Message msg = MapViewMessage.showDialogMessage(dialogId.shortValue());
		((SecondUGASPClient)client).getMapActivity().getMapViewHandler().sendMessage(msg);
	}
	
	
	public void updateDialogFrame(Integer textViewIndex, Short dialogId){
		Log.d(getClass().getSimpleName(), "--- RMAI called = updateDialogFrame "+dialogId);
		Message msg = MapViewMessage.updateDialogFrameMessage(textViewIndex.intValue(), dialogId.shortValue());
		((SecondUGASPClient)client).getMapActivity().getMapViewHandler().sendMessage(msg);
	}
	
	
	public void cancelDialogFrame(){
		Log.d(getClass().getSimpleName(), "--- RMAI called = cancelDialogFrame ");
		Message msg = MapViewMessage.cancelDialogFrameMessage();
		((SecondUGASPClient)client).getMapActivity().getMapViewHandler().sendMessage(msg);
	}
	
	/**
	 * 
	 * @param name
	 * @param description
	 * @param marker
	 * @param hiddenElementHashCode used to remove the hidden element from the map's overlays.
	 */
	public void catchHiddenElement(String name, String description, Integer marker, Integer hiddenElementHashCode){
		Log.d(getClass().getSimpleName(), "--- RMAI called = catchHiddenElement ");
		Message msg = MapViewMessage.catchHiddenElement(name,description,marker.intValue(),hiddenElementHashCode.intValue());
		((SecondUGASPClient)client).getMapActivity().getMapViewHandler().sendMessage(msg);
	}
	
	
	public void sendTransportPoints(Short pointsToSend){
		Log.d(getClass().getSimpleName(), "--- RMAI called = sendTransportPoints "+pointsToSend);
		Message msg1 = MapViewMessage.displayLongToastMessage("Vous avez reçu "+pointsToSend+" points transport !");
		Message msg2 = MapViewMessage.sendTransportPointsMessage(pointsToSend.shortValue());
		((SecondUGASPClient)client).getMapActivity().getMapViewHandler().sendMessage(msg1);
		((SecondUGASPClient)client).getMapActivity().getMapViewHandler().sendMessage(msg2);
	}
	
	
	private static class RMAIArraysUtil{
		
		
		static GeoPoint toGeoPoint(Object[] array){
			GeoPoint geoPoint = null;
			if(array != null && array.length == 2){
				Integer[] tab = new Integer[2];
				// latitude
				tab[0] = (Integer)array[0];
				// longitude
				tab[1] = (Integer)array[1];
				geoPoint = new GeoPoint(tab[0].intValue(), tab[1].intValue());
			}
			return geoPoint;
		}
		
		static GeoPoint[] toGeoPointArray(Object[] array){
			GeoPoint[] geoPointArray = null;
			if(array != null){
				geoPointArray = new GeoPoint[array.length];
				for (int i = 0; i < geoPointArray.length; i++) {
					geoPointArray[i] = toGeoPoint((Object[])array[i]);
				}
			}
			return geoPointArray;
		}
		
		
		static List<Short> toListShort(Object[] array){
			List<Short> res = null;
			if(array != null){
				res = new ArrayList<Short>();
				for (int i = 0; i < array.length; i++) {
					res.add((Short)array[i]);
				}
			}
			return res;
		}
		
	}

}
