package org.playonline.network.client.custom;

import java.util.Vector;

import org.playonline.network.client.RMAIClientManager;
import org.playonline.network.client.http.UGASPClientHttp;
import org.playonline.network.lobby.ApplicationInstanceInfos;
import org.playonline.ui.basic.map.PlayOnlineMapActivity;
import org.playonline.ui.basic.map.handler.MapViewMessage;

import android.os.Message;
import android.util.Log;


/**
 * This class allow communication between two GASP clients.
 * Depending to the constructor, the client can be the owner of the game instance, or just an actor.
 * If it is the owner, it will create the game instance and wait for another client to start it. 
 * Otherwise, it will just get the list of available application instances, and join the first one.
 * 
 * Note : 
 * SecondGASPClient should be launch typing following line :
 * mvn -DGASP-client=second -Dexec.args="X Y (Z)" install
 * with the arguments X, Y and possibly Z described in UGASPWhat() method javadoc. 
 */
public class SecondUGASPClient extends UGASPClientHttp implements Runnable {

	private boolean isCreator;
	/**
	 * Following values allow to succeed in login (for the owner, which already registered in uGASP's database)
	 * and firstLogin (for "simple" client).
	 */
	// Just used by the owner on the login
	public static final String OWNER_USERNAME  = "Agent X";
	private static final String OWNER_PASSWORD 	= "2";
	private static final String DEFAUT_PASSWORD = "furGFT54";
	
	// Used by simple clients on the first login		
	public static final int applicationID = 2;

	// For createAI(m_sessionId, minActors, maxActors)
	// Values are modifiable
	public static final int  MIN_ACTORS = 1;
	public static final int  MAX_ACTORS = 100;
	
	// In game parameters
	private String myPassword;

	private final PlayOnlineMapActivity mapActivity;
	
	
	// Loggin messages
	public static final String CREATE_INSTANCE_MESSAGE 			= "Game instance created.";
	public static final String CREATE_INSTANCE_FAILED_MESSAGE 	= "uGASPClient failed in creating a game instance";
	public static final String JOIN_INSTANCE_MESSAGE			= "Game instance joined";
	public static final String JOIN_INSTANCE_FAILED_MESSAGE 	= "uGASPClient failed in joining a game instance";
	
	
	private SecondUGASPClient(PlayOnlineMapActivity mapActivity){
		this.mapActivity = mapActivity;
	}

	/**
	 * Creates a SecondGASPClient that will be the owner of a game instance.
	 * 
	 * @param MAX_ACTORS the number of actors needed to start part
	 * @return SecondGASPClient the client which creates and start the game instance. 
	 */
	public static SecondUGASPClient getCreator(PlayOnlineMapActivity mapActivity){
		SecondUGASPClient owner = new SecondUGASPClient(mapActivity);
		owner.isCreator = true;
		owner.getID().setUsername(OWNER_USERNAME);
		owner.myPassword = OWNER_PASSWORD;
		return owner;
	}

	/**
	 * Create a SecondGASPClient that will try to join a game instance.
	 * @param username   the user name of the client
	 * @return SecondGASPClient the client which join the game instance. 
	 */
	public static SecondUGASPClient getSimpleClient(PlayOnlineMapActivity mapActivity,String username){
		SecondUGASPClient client = new SecondUGASPClient(mapActivity);
		client.isCreator = false;
		client.getID().setUsername(username);
		client.myPassword = DEFAUT_PASSWORD;
		return client;
	}

	/**
	 * Depending whether the client is the owner or not, this method:
	 * <ul>
	 * <li>create an application instance,</li>
	 * <li>join an application instance.</li>
	 * </ul>
	 * In both cases, if logging succeed, data transmission routine is launched.
	 *
	 */
	public void run(){
		// DEBUG only
		//long start = System.currentTimeMillis();
		//long end;
		Message createInstanceDialog, joinInstanceMessage, displayLongToastMessage, cancelMessage;
		if(isCreator){
			createInstanceDialog = MapViewMessage.displayCreateInstanceProgressDialog();
			mapActivity.getMapViewHandler().sendMessage(createInstanceDialog);
			short m_actorSessionID = createAndStartUGaspApplicationInstance();
			//end = System.currentTimeMillis();
			if(m_actorSessionID > 0){
				Log.i(this.getClass().getSimpleName(),CREATE_INSTANCE_MESSAGE);
				displayLongToastMessage = MapViewMessage.displayLongToastMessage(CREATE_INSTANCE_MESSAGE);
				mapActivity.getMapViewHandler().sendMessage(displayLongToastMessage);
			}else{
				Log.i(this.getClass().getSimpleName(),CREATE_INSTANCE_FAILED_MESSAGE);
				displayLongToastMessage = MapViewMessage.displayLongToastMessage(CREATE_INSTANCE_FAILED_MESSAGE);
				mapActivity.getMapViewHandler().sendMessage(displayLongToastMessage);
				cancelMessage = MapViewMessage.cancelProgressDialog();
				mapActivity.getMapViewHandler().sendMessage(cancelMessage);
				return;
			}
		} else{
			joinInstanceMessage = MapViewMessage.displayJoinInstanceProgressDialog();
			mapActivity.getMapViewHandler().sendMessage(joinInstanceMessage);
			int m_actorSessionID = joinUGaspApplicationInstance();
			//end = System.currentTimeMillis();
			if(m_actorSessionID > 0){
				Log.i(this.getClass().getSimpleName(),JOIN_INSTANCE_MESSAGE);
				displayLongToastMessage = MapViewMessage.displayLongToastMessage(JOIN_INSTANCE_MESSAGE);
				mapActivity.getMapViewHandler().sendMessage(displayLongToastMessage);
			}else{
				Log.i(this.getClass().getSimpleName(),JOIN_INSTANCE_FAILED_MESSAGE);
				displayLongToastMessage = MapViewMessage.displayLongToastMessage(JOIN_INSTANCE_FAILED_MESSAGE);
				mapActivity.getMapViewHandler().sendMessage(displayLongToastMessage);
				cancelMessage = MapViewMessage.cancelProgressDialog();
				mapActivity.getMapViewHandler().sendMessage(cancelMessage);
				return;
			}
		}
		cancelMessage = MapViewMessage.cancelProgressDialog();
		mapActivity.getMapViewHandler().sendMessage(cancelMessage);
		// DEBUG only
		//displayLongToastMessage = MapViewMessage.displayLongToastMessage("Logging duration = "+(end-start)+" ms.");
		//mapActivity.getMapViewHandler().sendMessage(displayLongToastMessage);
		// for every client, starts the client's routine
		getDataTransmissionMgr().launchDataTransmissionRoutine();
	}

	/*
	 * This method successively proceeds :
	 * - a login to get an sessionID
	 * - an application instance creation which return an actorSessionID.
	 *@return the actorSessionID of the owner
	 */
	private short createAndStartUGaspApplicationInstance(){
		// first login with static values of username and password
		getLobbyMgr().firstLogin(applicationID, getID().getUsername(), myPassword);
		if(getID().getActorID() > 0){
			// getUsername() and myPassword are equals respectively to ownerUsername and ownerPassword
			getLobbyMgr().login(getID().getActorID(), getID().getUsername(), myPassword);
			if(getID().getSessionID() > 0){
				getLobbyMgr().createApplicationInstance(getID().getSessionID(), MIN_ACTORS, MAX_ACTORS);
				if(getID().getActorSessionID() > 0){
					// pseudo is changed with user name value
					getLobbyMgr().changePseudo(getID().getActorSessionID(), getID().getUsername());
					return getID().getActorSessionID();	
				} else{
					Log.i(this.getClass().getSimpleName(),"Failed in creating application instance");
					return -1;
				}
			}else {
				Log.i(this.getClass().getSimpleName(),"Failed in login");
				return -1;
			}
		}else{
			Log.i(this.getClass().getSimpleName(),"Failed in first login");
				return -1;
		}
	}

	/*
	 * This method successively proceeds :
	 * - a first login to get an actorID
	 * - a login to get an sessionID
	 * - an application instance creation which return an actorSessionID.
	 *@ return the actorSessionID of the client
	 */
	private short joinUGaspApplicationInstance(){
		getLobbyMgr().firstLogin(applicationID, getID().getUsername(), myPassword);
		if(getID().getActorID() > 0){
			getLobbyMgr().login(getID().getActorID(), getID().getUsername(), myPassword);
			if(getID().getSessionID()> 0){
				// get the list of available game instances
				Vector<ApplicationInstanceInfos> applicationInstances = getLobbyMgr().getApplicationInstancesInfos(getID().getSessionID());
				if(applicationInstances.size()!=0){
					getLobbyMgr().joinApplicationInstance(getID().getSessionID(), applicationInstances.elementAt(0).getApplicationInstanceID());
					if(getID().getActorSessionID() > 0){
						// pseudo is changed with user name value
						getLobbyMgr().changePseudo(getID().getActorSessionID(), getID().getUsername());
						return getID().getActorSessionID();	
					} else{
						Log.i(this.getClass().getSimpleName(),"Failed in creating application instance");
						return -1;
					}
				}else{
					Log.i(this.getClass().getSimpleName(),"No available application instance");
					return -1;
				}
			}else {
				Log.i(this.getClass().getSimpleName(),"Failed in login");
				return -1;
			}
		}else{
			Log.i(this.getClass().getSimpleName(),"Failed in first login");
			return -1;
		}
	}
	
	@Override
	public void onEndEvent(short actorSessionId) {
		Message msg = MapViewMessage.displayAlertMessage("Application Instance is ended.");
		mapActivity.getMapViewHandler().sendMessage(msg);
	}
	
	@Override
	public void onQuitEvent(short actorSessionId) {
		Message msgToast = MapViewMessage.displayLongToastMessage("ActorSessionId = "+actorSessionId+" quit the instance.");
		mapActivity.getMapViewHandler().sendMessage(msgToast);
		mapActivity.getCharactersManager().removedCharacter(Short.valueOf(actorSessionId));
	}
	
	@Override
	public void onStartEvent(short actorSessionId) {
		Message msg = MapViewMessage.displayAlertMessage("Application Instance is started.");
		mapActivity.getMapViewHandler().sendMessage(msg);
	}
	
	@Override
	public void onJoinEvent(short actorSessionId, String username) {
		// if client's own joint event is received
		if(actorSessionId == getID().getActorSessionID()){
			// query is sent to platform to retrieve informations about other characters
			getDataTransmissionMgr().getRMAIMgr().remoteMethodAsynchronousInvocation(
					RMAIClientManager.ActorSessionID.uGASP_GAME_LOGIC.getID(), 
					"retrievedCharactersInformations");
			// location is sent to platform
			getDataTransmissionMgr().getRMAIMgr().remoteMethodAsynchronousInvocation(
					RMAIClientManager.ActorSessionID.uGASP_GAME_LOGIC.getID(), 
					"storeCharacter",
					new Integer(mapActivity.getCharactersManager().getMyCharacterLocation().getLatitudeE6()),
					new Integer(mapActivity.getCharactersManager().getMyCharacterLocation().getLongitudeE6()),
					mapActivity.getCharacter().getName());
		}//otherwise log message is displayed to inform new player presence
		else{
			Message msgToast = MapViewMessage.displayLongToastMessage(username+" with actorSessionID = "+actorSessionId+" join the instance.");
			mapActivity.getMapViewHandler().sendMessage(msgToast);
		}
	}

	
	public PlayOnlineMapActivity getMapActivity(){
		return mapActivity;
	}
	
	
	public boolean isCreator(){
		return isCreator;
	}
}