package org.playonline.network.connectedclients;

import java.util.List;
import java.util.Vector;

public class ConnectedClientMgr {

	private final List<ConnectedClient> connectedClients;// Vector where connected clients are stored
	
	public ConnectedClientMgr(){
		connectedClients = new Vector<ConnectedClient>();
	}
	
	/**
	 * Insert a connected client into the vector of connected clients, if its not present yet.
	 *
	 *@return	true if the client has been inserted into the vector of connected clients,
	 *			false otherwise.
	 **/
	public boolean addConnectedClient(ConnectedClient client){
		// get a lock on the vector of connected clients
		synchronized (connectedClients) {
			if(!connectedClients.contains(client)){
				connectedClients.add(client);
				return true;
			}else
				// if client is already in the vector of connected clients
				return false;
		}	
	}

	/**
	 * Check whether a client is present in the vector of connected clients.
	 * If so, the client is removed from the vector.
	 *
	 *@return    true if the client matching with the actorSessionID has been removed,
	 *			false otherwise.
	 **/
	public boolean removeConnectedClient(int actorSessionID){
		// get a lock on the vector of connected clients
		synchronized (connectedClients) {
			ConnectedClient clientToRemove = getConnectedClientByActorSessionID(actorSessionID);
			if(!clientToRemove.equals(null)){
				connectedClients.remove(clientToRemove);
				return true;
			}else
				// if client isn't in the vector of connected clients
				return false;
		}
	}
	
	public void clearConnectedClientsList(){
		// get a lock on the vector of connected clients
		synchronized (connectedClients) {
			connectedClients.clear();
		}
	}

	/**
	 *	Get a ConnectedClient from the connectedClients vector with its actorSessionID.
	 *
	 *@param 	actorSessionID	the actorSessionID of the requested client
	 *@return 	ConnectedClient if the matching actorSessionID is found in connectedClients,
	 *			null otherwise.
	 **/
	public ConnectedClient getConnectedClientByActorSessionID(int actorSessionID){
		// get a lock on the vector of connected clients
		synchronized (connectedClients) {
			for(int i = 0; i < connectedClients.size(); i++){
				if (((ConnectedClient)((Vector<ConnectedClient>) connectedClients).elementAt(i)).getActorSessionID() == actorSessionID)
					return (ConnectedClient)((Vector<ConnectedClient>) connectedClients).elementAt(i);
			}
		}
		// if client isn't in the vector of connected clients
		return null;
	}


	/**
	 *	Get a copy of the vector of the connected clients.
	 *	
	 *@return	A clone of the connectedClients vector
	 **/
	public Vector<ConnectedClient> getConnectedClientsList(){
		// get a lock on the vector of connected clients
		synchronized (connectedClients) {
			// Need to be cast first in Vector to enable use of clone method
			return (Vector<ConnectedClient>) (((Vector) connectedClients).clone());
		}
	}

	/**
	 *	Get the size of the vector of the connected clients.
	 *	
	 *@return	the size of the connectedClients vector
	 **/
	public int getNumberOfConnectedClients(){
		// get a lock on the vector of connected clients
		synchronized (connectedClients) {
			return connectedClients.size();
		}
	}
}
