package org.playonline.network.datatransmission.events;

import java.io.DataInputStream;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;

import org.playonline.network.client.RMAIClientManager;
import org.playonline.network.connectedclients.ConnectedClient;
import org.playonline.network.connectedclients.ConnectedClientMgr;
import org.playonline.network.datatransmission.DataTransmissionMgr;
import org.playonline.network.datatransmission.protocol.rmai.tools.InvokedMethod;
import org.ugasp.eventmodel.event.impl.EventsType;

public class EventsMgrImpl implements EventsMgr{

	private final DataTransmissionMgr	dataTransmissionMgr;
	private final ConnectedClientMgr 	connectedClientMgr;
	
	public EventsMgrImpl(DataTransmissionMgr dataTransmissionMgr){
		this.dataTransmissionMgr	= dataTransmissionMgr;
		this.connectedClientMgr 	= new ConnectedClientMgr();
	}
	
	/**
	 * Treat the events of the servlet response data input stream.
	 * Except for the StartEvent, every other event received triggers a call to an intermediate method
	 * (onJoin, onQuit, onEnd or onData), before calling the onJoinEvent, onEndEvent, onQuitEvent or onDataEvent methods.
	 * 
	 * @param dis the data input stream
	 * @throws IOException
	 */
	public void eventsOf(DataInputStream dis) throws IOException {
		byte type;
		// as long as data are in the input stream
		while(dis.available()>0){
			type = dis.readByte(); // get the type of the events

			switch (type){
			case EventsType.JOIN_EVENT_TYPE:
				onJoin(dis.readShort(),dis.readUTF());
				break;
			case EventsType.START_EVENT_TYPE:
				dataTransmissionMgr.getUGASPClient().onStartEvent(dis.readShort());
				break;
			case EventsType.END_EVENT_TYPE:
				onEnd(dis.readShort());
				break;
			case EventsType.QUIT_EVENT_TYPE:
				onQuit(dis.readShort());
				break;
			case EventsType.RMAI_EVENT_TYPE:
				// retrieve sender, recipient, and the string representing XML serialization of invoked method
				onRMAI(dis.readShort(),dis.readShort(),dis.readUTF()); 
				break;
			default:  break;
			}
		}
	}
	
	
	/*
	 * Store a client in the vector of connected clients if it wasn't present yet, 
	 * and call the onJoinEvent(int actorSessionID, String username) method.
	 *
	 *@param actorSessionID	the actorSessionID of the client to store
	 *@param username		its username
	 */
	private void onJoin(short actorSessionID, String username){
		ConnectedClient client = new ConnectedClient(actorSessionID,username);
		// add the client if it is not present in the vector of connected clients
		connectedClientMgr.addConnectedClient(client);
		// call the onJoinEvent method
		dataTransmissionMgr.getUGASPClient().onJoinEvent(actorSessionID,username);	
	}

	/*
	 * Remove a client from the vector of connected clients if it was present, 
	 * and call the onQuitEvent(int actorSessionID) method.
	 *
	 *@param actorSessionID	the actorSessionID of the client to store
	 */
	private void onQuit(short actorSessionID){
		connectedClientMgr.removeConnectedClient(actorSessionID);
		// call the onQuitEvent method
		dataTransmissionMgr.getUGASPClient().onQuitEvent(actorSessionID);	
	}

	/*
	 * This method:
	 * - clears the vector of connected clients,
	 * - sets the instanceQuit variable of the Lobby Mgr of UGASPClient to true (--> stop the data transmission routine),
	 * - invalidates the actor session ID
	 * - calls the onEndEvent(short actorSessionID) method.
	 *
	 *@param actorSessionID	the actorSessionID of the client which launched the EndEvent
	 */
	private void onEnd(short actorSessionID){
		connectedClientMgr.clearConnectedClientsList();
		// stop the data transmission routine
		dataTransmissionMgr.getUGASPClient().getLobbyMgr().setInstanceQuit(true);
		// actor session ID is set to an invalid value
		dataTransmissionMgr.getUGASPClient().getID().invalidateActorSessionID();
		// call the onQuitEvent method
		dataTransmissionMgr.getUGASPClient().onEndEvent(actorSessionID);	
	}


	private void onRMAI(short senderActorSessionID, short recipientActorSessionID, String invokedMethodXMLSerialized){
		// declare InvokedMethod and Method needed to execute the RMAI
		Method rmaiMethod;
		InvokedMethod toInvoke;
		// create a RMAIClientManager
		RMAIClientManager rmaiClientManager = new RMAIClientManager(dataTransmissionMgr.getUGASPClient(),senderActorSessionID,recipientActorSessionID);
		try {
			// de-serialize the XML
			toInvoke = dataTransmissionMgr.getRMAIMgr().getRmaiUtils().readMethodInvocation(invokedMethodXMLSerialized);
			// get the class of the attribute on which method should be invoked, i.e. ClientManager class
			Class<?> rmaiClientManagerClass = rmaiClientManager.getClass();
			// get a Method object matching with informations included in toInvoke ...
			rmaiMethod = rmaiClientManagerClass.getMethod(toInvoke.getMethodName(),toInvoke.getParametersTypes());
			// invoke the method on the rmai client manager with the suitable arguments
			rmaiMethod.invoke(rmaiClientManager, toInvoke.getParameters());
		} catch (ClassNotFoundException e) {
			e.printStackTrace();
		} catch (InstantiationException e) {
			e.printStackTrace();
		} catch (NoSuchMethodException e) {
			e.printStackTrace();
		} catch (IllegalAccessException e) {
			e.printStackTrace();
		} catch (InvocationTargetException e) {
			e.printStackTrace();
		} catch (SecurityException e) {
			e.printStackTrace();
		}
	}
}
