package org.playonline.network.datatransmission.http;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.URL;

import org.playonline.network.client.UGASPClient;
import org.playonline.network.client.custom.SecondUGASPClient;
import org.playonline.network.client.http.UGASPClientHttp;
import org.playonline.network.datatransmission.DataTransmissionMgr;
import org.playonline.network.datatransmission.events.EventsMgr;
import org.playonline.network.datatransmission.events.EventsMgrImpl;
import org.playonline.network.datatransmission.protocol.exception.DataTransmissionProtocolVersionException;
import org.playonline.network.datatransmission.protocol.rmai.RMAIMgr;
import org.playonline.network.datatransmission.protocol.rmai.RMAIMgrImpl;
import org.playonline.ui.basic.map.handler.MapViewMessage;

import android.os.Message;
import android.util.Log;

public class DataTransmissionMgrHttp implements DataTransmissionMgr {

	private static long 		period = 1000;// Time period between each getEvents, in ms
	private static final String DATA_TRANSMISSION_SERVLET_PATH = "/DataTransmissionServlet";// servlet dedicated to data transmission
	private static URL 			dataTransmissionServletURL= getDataTransmissionServletURL();
	private static final byte 	DATA_TRANSMISSION_PROTOCOL_VERSION = 1;
	private static final byte 	PLAYER_REQUESTER_TYPE = 2;
	private static final byte 	GET_EVENTS_REQUESTER_TYPE = 0;
	// Data transmission header size = 6 + GET_EVENT_TYPE size = 1
	private static final int 	BASIC_DATA_TRANMISSION_MESSAGE_LENGTH = 7;
	
	// boolean used to stop client in case of disconnection
	protected boolean 			disconnected = false;
	private HttpURLConnection dataTransmissionConnection;

	private ByteArrayOutputStream storageBufferByteArrayOutputStream, sendBufferByteArrayOutputStream;
	private DataOutputStream storageBufferOutputStream;

	private byte messageSentID;
	private byte lastMessageReceivedID;

	
	private final UGASPClient 	uGASPClient;
	private final RMAIMgr 		rmaiMgr;
	private final EventsMgr 	eventsMgr;
	
	public DataTransmissionMgrHttp(UGASPClient uGaspClient){
		this.uGASPClient 	= uGaspClient;
		this.rmaiMgr		= new RMAIMgrImpl(this);
		this.eventsMgr 		= new EventsMgrImpl(this);
	}
	
	public UGASPClient getUGASPClient() {
		return uGASPClient;
	}

	public RMAIMgr getRMAIMgr() {
		return rmaiMgr;
	}

	public EventsMgr getEventsMgr() {
		return eventsMgr;
	}
	
	public void setDisconnected(boolean disconnected) {
		this.disconnected = disconnected;
	}
	
	public DataOutputStream getStorageBufferOutputStream() {
		return storageBufferOutputStream;
	}

	public ByteArrayOutputStream getStorageBufferByteArrayOutputStream() {
		return storageBufferByteArrayOutputStream;
	}
	
	public void reinitializeStorageBuffer() {
		storageBufferOutputStream = new DataOutputStream(storageBufferByteArrayOutputStream);
	}
	
	public void resetStorageBuffer() {
		try {
			// flush storage buffer output stream
			storageBufferOutputStream.flush();
			// reset storage buffer
			storageBufferByteArrayOutputStream.reset();
		} catch (Throwable t2) {
			Log.w(getClass().getSimpleName(), "Exception on flushing of storage buffer",t2);
			// reinitialize storage buffer
			storageBufferOutputStream = new DataOutputStream(storageBufferByteArrayOutputStream);
		}
	}

	private final static URL getDataTransmissionServletURL(){
		URL url = null;
		try {
			url = new URL(UGASPClientHttp.uGASPServletUrl+DATA_TRANSMISSION_SERVLET_PATH);
		} catch (MalformedURLException e) {
			e.printStackTrace();
		} 
		return url;
	}
	
	public static void updateUGASPUrl(){
		dataTransmissionServletURL= getDataTransmissionServletURL();
	}
	
	/** 
	 *  This method represents the routine for client when 
	 *  a game instance is started : 
	 *  <ul>
	 *  <li>client send RMAI and/or MooDS objects if there are some</li>
	 *  <li>client is getting events from platform if there are some</li>
	 *  <li>then, it sleeps for a time interval corresponding to GASPClient.period.</li>
	 *  </ul>
	 *  Method stops as soon as :
	 *  <ul>
	 *  <li>QuitApplicationInstance, EndApplicationInstance or QuitPlatform have been called</li>
	 *  <li>a disconnection occurs.</li>
	 *  </ul> 
	 */
	public void launchDataTransmissionRoutine() {
		// check whether actor session ID is valid
		if(uGASPClient.getID().getActorSessionID()>0){
			// initialize attributes
			initTransmissionsAttributes();
			DataOutputStream dataTransmissionOutputStream = null;
			DataInputStream dataTransmissionInputStream = null;
			// for test
			boolean connected = true;
			/****************************************************************/
			/************************** routine *****************************/
			/****************************************************************/
			while(!uGASPClient.getLobbyMgr().isInstanceQuit() && !disconnected){
				try{
					/******************************************************************************/
					/************************** Sending data to uGASP *****************************/
					/******************************************************************************/
					// initialize data transmission connection and gets its data output stream
					dataTransmissionOutputStream = initDataTransmissionConnection();
					if(dataTransmissionOutputStream != null){
						// if uGASP hasn't acknowledged last sending of data --> old data (and new if so) are sent back to uGASP
						if(sendBufferByteArrayOutputStream.size() != 0){
							Log.v(getClass().getSimpleName(), "Data Transmission routine: sending back un-acknoledged data to uGASP platform.");
							// if storage buffer isn't empty, it is copied at the end of send buffer
							copyStorageBufferIntoSendBufferIfNeeded();
							writeDataOnTransmissionOutputStream(dataTransmissionOutputStream, uGASPClient.getID().getActorSessionID());
						}
						// if uGASP has acknowledged last sending of data
						else{
							// if storage buffer isn't empty, it is copied into send buffer
							copyStorageBufferIntoSendBufferIfNeeded();
							// test whether client has called one of the RMAI methods or sendMooDSObject during its period
							if(sendBufferByteArrayOutputStream.size() != 0){
								Log.v(getClass().getSimpleName(), "Data Transmission routine: RMAI are sent.");
								writeDataOnTransmissionOutputStream(dataTransmissionOutputStream, uGASPClient.getID().getActorSessionID());
							}
							// if client hasn't any data to send to uGASP
							else{
								Log.v(getClass().getSimpleName(), "Data Transmission routine: empty message sent to uGASP.");
								// length of a "simple" get events request is BASIC_DATA_TRANMISSION_MESSAGE_LENGTH
								dataTransmissionOutputStream = writeRequestHeaderOnDataTransmissionOutputStream(dataTransmissionOutputStream, BASIC_DATA_TRANMISSION_MESSAGE_LENGTH, uGASPClient.getID().getActorSessionID(), messageSentID, lastMessageReceivedID);
								// finish request by a GET_EVENT_REQUESTER_TYPE
								dataTransmissionOutputStream.writeByte(GET_EVENTS_REQUESTER_TYPE);
							}
						}
						// in every cases
						dataTransmissionOutputStream.close();						
						/***************************************************************************************/
						/************************** Reception of events from uGASP *****************************/
						/***************************************************************************************/			

						// return connection's data input stream if response code is OK 
						if (dataTransmissionConnection.getResponseCode() == HttpURLConnection.HTTP_OK){
							// for test
							if(!connected){
								Message msg = MapViewMessage.displayShortToastMessage("Reconnected");
								((SecondUGASPClient)getUGASPClient()).getMapActivity().getMapViewHandler().sendMessage(msg);
								connected = true;
							}
							boolean treated = treatUGASPResponse(dataTransmissionInputStream);
							if(!treated){
								Log.w(getClass().getSimpleName(), "Invalid response from uGASP received --> an error has possibly occured on client message transmission.");
							}
						}else{
							Log.w(getClass().getSimpleName(),"HTTP response from uGASP invalid, code = "+dataTransmissionConnection.getResponseCode());
						}
						/***************************************************************************************/
						/************************** Sleep until next period ************************************/
						/***************************************************************************************/
						sleepUntilNextPeriod();
					}
					// connection output stream is null
					else{
						// TODO try to reconnect ? increment a count ?
					}
				}catch(DataTransmissionProtocolVersionException e){
					Log.e(this.getClass().getSimpleName(), e.getMessage());
					// quit the application instance
					uGASPClient.getLobbyMgr().quitApplicationInstance(uGASPClient.getID().getActorSessionID());
				}catch(Throwable t){
					Log.w(this.getClass().getSimpleName(), "Exception occured during Data Transmission routine.",t);
					// for test 
					connected = false;
					Message msg = MapViewMessage.displayShortToastMessage("Disconnected");
					((SecondUGASPClient)getUGASPClient()).getMapActivity().getMapViewHandler().sendMessage(msg);
					//TODO Display + newRun + count incremented ?
					uGASPClient.handleDisconnection();
				}
			}
		}else{
			Log.w(getClass().getSimpleName(),"Data Transmission Routine cannot be lauched until uGASP client has not retrieved a valid Actor Session ID");
		}
	}
	
	/*
	 * Initialize instance attributes used by data transmission protocol.
	 */
	private void initTransmissionsAttributes(){
		storageBufferByteArrayOutputStream= new ByteArrayOutputStream();
		storageBufferOutputStream = new DataOutputStream(storageBufferByteArrayOutputStream);
		sendBufferByteArrayOutputStream = new ByteArrayOutputStream();
		messageSentID = 0;
		lastMessageReceivedID = Byte.MIN_VALUE;
	}
	
	/*
	 * As we cannot synchronized on connections and data streams since they are not instantiate,
	 * this method is declared synchronized.
	 * This method initiates the dataTransmissionConnection.
	 * TODO How should we properly synchronized connection ?
	 */
	private synchronized DataOutputStream initDataTransmissionConnection() throws IOException{
		DataOutputStream dos = null;	
		// open the connection matching with this URL
		dataTransmissionConnection = (HttpURLConnection) dataTransmissionServletURL.openConnection();
		// setting connection
		dataTransmissionConnection.setDoOutput(true);
		dataTransmissionConnection.setRequestMethod("POST");
		dataTransmissionConnection.setRequestProperty("Content-Type", "application/octet-stream");
		// get DataOutputStream from the connection
		dos = new DataOutputStream(dataTransmissionConnection.getOutputStream());		
		return dos;
	}
	
	/*
	 * Test whether storageBuffer has any data. 
	 * If so, these data are copied at the end of sendBuffer,
	 * then storageBuffer is flushed.
	 * 
	 * return false whether storage buffer or send buffer is null
	 */
	//TODO Insert buffer permutation here if storage Buff isn't empty and send Buff is empty
	private boolean copyStorageBufferIntoSendBufferIfNeeded() throws IOException{
		if(storageBufferByteArrayOutputStream != null && sendBufferByteArrayOutputStream != null){
			// get a lock on the storageBufferByteArrayOutputStream
			synchronized(storageBufferByteArrayOutputStream){
				// test whether client has called one of the RMAI methods or sendMooDSObject during its period
				if (storageBufferByteArrayOutputStream.size() != 0){
					Log.v(getClass().getSimpleName(), "Storage buffer size = "+storageBufferByteArrayOutputStream.size());
					// copy data contained in storageBuffer at the end of sendBuffer
					storageBufferByteArrayOutputStream.writeTo(sendBufferByteArrayOutputStream);
					// reset storageBuffer
					storageBufferByteArrayOutputStream.reset();
				}
			}
			return true;
		}else{
			return false;
		}
	}
	
	/*
	 * This method successively
	 * <ul>
	 * <li>write suitable header on the data transmission output stream</li>
	 * <li>write data to transmit</li>
	 * <li>write a GET_EVENTS_REQUESTER_TYPE</li>
	 * </ul>
	 * return false if dataTransmissionOutputStream is null
	 */
	private boolean writeDataOnTransmissionOutputStream(DataOutputStream dataTransmissionOutputStream, short actorSessionID) throws IOException{
		if(dataTransmissionOutputStream != null){
			// sendBufferByteArrayOutputStream.size()+1 because we finish by a GET_EVENTS_REQUESTER_TYPE
			dataTransmissionOutputStream = writeRequestHeaderOnDataTransmissionOutputStream(dataTransmissionOutputStream, BASIC_DATA_TRANMISSION_MESSAGE_LENGTH + sendBufferByteArrayOutputStream.size(), actorSessionID, messageSentID, lastMessageReceivedID);
			// write RMAI & MooDS operation executed by client during last period, and possibly during this period, if so
			dataTransmissionOutputStream.write(sendBufferByteArrayOutputStream.toByteArray());
			// finish request by a GET_EVENT_REQUESTER_TYPE
			dataTransmissionOutputStream.writeByte(GET_EVENTS_REQUESTER_TYPE);
			Log.d(getClass().getSimpleName(), "--- dataTransmissionOutputStream  size = "+dataTransmissionOutputStream.size());
			return true;
		}else{
			return false;
		}
	}
	
	
	/*
	 * DataTransmissionOutputStream must be initialized with initDataTansmissionConnection()
	 * before calling this method
	 * TODO Otherwise ?? 
	 */
	private DataOutputStream writeRequestHeaderOnDataTransmissionOutputStream(DataOutputStream dataTransmissionOutputStream, int messageLength, short actorSessionID, byte messageSentID, byte lastMessageReceivedID) throws IOException{
		dataTransmissionOutputStream.writeInt(messageLength);
		dataTransmissionOutputStream.writeByte(DATA_TRANSMISSION_PROTOCOL_VERSION);
		dataTransmissionOutputStream.writeByte(PLAYER_REQUESTER_TYPE);
		dataTransmissionOutputStream.writeShort(actorSessionID);
		dataTransmissionOutputStream.writeByte(messageSentID);
		dataTransmissionOutputStream.writeByte(lastMessageReceivedID);
		return dataTransmissionOutputStream;
	}
	
	
	/*
	 * This method successively
	 * <ul>
	 * <li>retrieve input stream from data transmission connection</li>
	 * <li>check header of this input stream</li>
	 * <li>treat events included in the input stream, if so</li>
	 * </ul>
	 * return false if response from uGASP is invalid or if dataTransmissionInputStream is null
	 */
	private boolean treatUGASPResponse(DataInputStream dataTransmissionInputStream) throws IOException, DataTransmissionProtocolVersionException{
		if(dataTransmissionConnection != null){
			// get events from the server into an inputStream
			dataTransmissionInputStream =  new DataInputStream(dataTransmissionConnection.getInputStream());
			if(dataTransmissionInputStream.available() != 0){
				// stream used to retrieve whole data inside dataTransmissionInputStream
				DataInputStream copiedInputStream = null;
				// copy dataTransmissionInputStream data into copiedInputStream
				byte[] data = new byte[dataTransmissionInputStream.available()];
				dataTransmissionInputStream.readFully(data);
				ByteArrayInputStream dataInputStream = new ByteArrayInputStream(data);
				copiedInputStream = new DataInputStream(dataInputStream);
				// check the header of the response
				copiedInputStream = checkUGASPResponse(copiedInputStream);
				// if response from uGASP is invalid
				if(copiedInputStream == null){
					return false;
				}
				// read events from the copyInputStream
				eventsMgr.eventsOf(copiedInputStream);
				// close input streams
				dataTransmissionInputStream.close();
				copiedInputStream.close();	
			}else{
				Log.d(getClass().getSimpleName(), "Response from uGASP is empty.");
			}
			// in both cases
			return true;
		}else{
			return false;
		}
	}
	
	
	/*
	 * DataTransmissionInputStream must be initialized before calling this method
	 */
	private DataInputStream checkUGASPResponse(DataInputStream dataTransmissionInputStream) throws IOException, DataTransmissionProtocolVersionException{
		// if uGASP response is invalid
		if(dataTransmissionInputStream.available() == 0){
			return null;
		}
		// retrieve the length of the message sent by uGASP
		int messageLength = dataTransmissionInputStream.readInt();
		if(messageLength != dataTransmissionInputStream.available()){
			Log.w(getClass().getSimpleName(), 	"uGASP Message is ignored, since excepted message length = "+
												messageLength+
												" doesn't match with available bits = "+
												dataTransmissionInputStream.available());
			return null;
		}
		byte protocolVersion = dataTransmissionInputStream.readByte();
		// test whether protocol version is the proper one
		if(protocolVersion < DATA_TRANSMISSION_PROTOCOL_VERSION){
			throw new DataTransmissionProtocolVersionException(DATA_TRANSMISSION_PROTOCOL_VERSION,protocolVersion);
		}
		// Read bytes before treating them
		byte receivedMessageID = dataTransmissionInputStream.readByte();
		byte receivedAckMessageID = dataTransmissionInputStream.readByte();
		// test whether platform has acknowledged message
		if(receivedAckMessageID == (messageSentID)){
			Log.v(getClass().getSimpleName(), 	"uGASP has acknoledged the message with ID = "+messageSentID);
			// sendBufferByteArrayOutputStream can hence be reset
			sendBufferByteArrayOutputStream.reset();
			// update message ID (modulo is used to always deal with positive values of ID, negative ones are used for initialization)
			messageSentID = (byte)((messageSentID + 1)%Byte.MAX_VALUE);
		}else{
			Log.d(getClass().getSimpleName(), 	"uGASP has NOT acknoledged the message with ID = "+messageSentID);
		}
		// update ackMessageID
		lastMessageReceivedID = receivedMessageID;
		return dataTransmissionInputStream;
	}
	
	/*
	 * Sleep for a duration defined by static attribute period
	 */
	private void sleepUntilNextPeriod(){
		try {
			Thread.sleep(period);
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
	}

}
