package org.playonline.network.datatransmission.protocol.rmai;

import java.io.ByteArrayOutputStream;
import java.io.DataOutputStream;

import org.playonline.network.datatransmission.DataTransmissionMgr;
import org.playonline.network.datatransmission.protocol.rmai.tools.RMAIUtils;

import android.util.Log;

public class RMAIMgrImpl implements RMAIMgr {

	private static final byte 	RMAI_REQUESTER_TYPE = 1;
	
	private final 	DataTransmissionMgr dataTransmissionMgr;;
	private final 	RMAIUtils 			rmaiUtils;// used to serialize and parse RMAI in a XML way
	
	public RMAIMgrImpl(DataTransmissionMgr dataMgr){
		this.dataTransmissionMgr 	= dataMgr;
		rmaiUtils 					= new RMAIUtils();
	}
	
	
	public final RMAIUtils getRmaiUtils() {
		return rmaiUtils;
	}

	
	/**
	 * Invoke asynchronously a remote method without parameters either :
	 * <ul>
	 * <li>on a uGASP Client, designated by its actor session id</li>
	 * <li>or on the uGASP game logic (RMAIClientManager.ActorSessionID.uGASP_GAME_LOGIC.getID())</li>
	 * <li>or on every connected uGASP clients (RMAIClientManager.ActorSessionID.BROADCAST_WITHOUT_GAME_LOGIC.getID())</li>
	 * <li>or on every connected uGASP clients and uGASP game logic (RMAIClientManager.ActorSessionID.BROADCAST_WITH_GAME_LOGIC.getID())</li>
	 * </ul>
	 * Methods that should be invoked must be implemented in the class RMAI client manager.
	 * Only instances methods can be invoked.
	 * 
	 * @param recipientActorSessionID	the actorSession ID of the suitable recipient
	 * @param methodName				the name of the method that should be invoked
	 * @return boolean					true if the response code of the HTTP connection is OK
	 * 									false otherwise
	 */
	public void remoteMethodAsynchronousInvocation(short recipientClientActorSessionID, String methodName) {
		if(dataTransmissionMgr.getUGASPClient().getID().isActorSessionIDValid()){
			// store an RMAI on the bufferOutputStream (third argument null means that there isn't any parameter in this RMAI)
			storeRMAI(recipientClientActorSessionID, methodName, null);
		}else{
			Log.w(getClass().getSimpleName(), "RMAI could not be invoked untill client has not retrieved a valid Actor Session ID");
		}

	}

	/**
	 * Invoke asynchronously a remote method either :
	 * <ul>
	 * <li>on a uGASP Client, designated by its actor session id</li>
	 * <li>or on the uGASP game logic (RMAIClientManager.ActorSessionID.uGASP_GAME_LOGIC.getID())</li>
	 * <li>or on every connected uGASP clients (RMAIClientManager.ActorSessionID.BROADCAST_WITHOUT_GAME_LOGIC.getID())</li>
	 * <li>or on every connected uGASP clients and uGASP game logic (RMAIClientManager.ActorSessionID.BROADCAST_WITH_GAME_LOGIC.getID())</li>
	 * </ul>
	 * Methods that should be invoked must be implemented in the class RMAI client manager.
	 * Only instances methods can be invoked.
	 * Here are the types that can be used as parameter of this remote method:
	 * <ul>
	 * <li>Boolean</li>
	 * <li>Byte</li>
	 * <li>Short</li>
	 * <li>Integer</li>
	 * <li>Long</li>
	 * <li>Double</li>
	 * <li>Float</li>
	 * <li>String</li>
	 * <li>RMAISerializable (you should pay a special attention on recursion : never pass this object reference).</li>
	 * <li>Arrays of those several types</li>
	 * </ul>
	 * 
	 * @param recipientActorSessionID	the actorSession ID of the suitable recipient
	 * @param methodName				the name of the method that should be invoked
	 * @param param0					first parameter of the method that should be invoked
	 * @return boolean					true if the response code of the HTTP connection is OK
	 * 									false otherwise
	 */
	public void remoteMethodAsynchronousInvocation(short recipientClientActorSessionID, String methodName, Object param0) {
		if(dataTransmissionMgr.getUGASPClient().getID().isActorSessionIDValid()){
			// build the Object array
			Object[] params = {param0};
			// store an RMAI on the bufferOutputStream
			storeRMAI(recipientClientActorSessionID, methodName, params);
		}else{
			Log.w(getClass().getSimpleName(), "RMAI could not be invoked untill client has not retrieved a valid Actor Session ID");
		}
	}

	/**
	 * Invoke asynchronously a remote method either :
	 * <ul>
	 * <li>on a uGASP Client, designated by its actor session id</li>
	 * <li>or on the uGASP game logic (RMAIClientManager.ActorSessionID.uGASP_GAME_LOGIC.getID())</li>
	 * <li>or on every connected uGASP clients (RMAIClientManager.ActorSessionID.BROADCAST_WITHOUT_GAME_LOGIC.getID())</li>
	 * <li>or on every connected uGASP clients and uGASP game logic (RMAIClientManager.ActorSessionID.BROADCAST_WITH_GAME_LOGIC.getID())</li>
	 * </ul>
	 * Methods that should be invoked must be implemented in the class RMAI client manager.
	 * Only instances methods can be invoked.
	 * Here are the types that can be used as parameter of this remote method:
	 * <ul>
	 * <li>Boolean</li>
	 * <li>Byte</li>
	 * <li>Short</li>
	 * <li>Integer</li>
	 * <li>Long</li>
	 * <li>Double</li>
	 * <li>Float</li>
	 * <li>String</li>
	 * <li>RMAISerializable (you should pay a special attention on recursion : never pass this object reference).</li>
	 * <li>Arrays of those several types</li>
	 * </ul>
	 * 
	 * @param recipientActorSessionID	the actorSession ID of the suitable recipient
	 * @param methodName				the name of the method that should be invoked
	 * @param param0					first parameter of the method that should be invoked
	 * @param param1					second parameter of the method that should be invoked
	 * @return boolean					true if the response code of the HTTP connection is OK
	 * 									false otherwise
	 */
	public void remoteMethodAsynchronousInvocation(
			short recipientClientActorSessionID, String methodName,
			Object param0, Object param1) {
		if(dataTransmissionMgr.getUGASPClient().getID().isActorSessionIDValid()){
				// build the Object array
				Object[] params = {param0,param1};
				// store an RMAI on the bufferOutputStream
				storeRMAI(recipientClientActorSessionID, methodName, params);	
			}else{
				Log.w(getClass().getSimpleName(), "RMAI could not be invoked untill client has not retrieved a valid Actor Session ID");
			}

	}

	/**
	 * Invoke asynchronously a remote method either :
	 * <ul>
	 * <li>on a uGASP Client, designated by its actor session id</li>
	 * <li>or on the uGASP game logic (RMAIClientManager.ActorSessionID.uGASP_GAME_LOGIC.getID())</li>
	 * <li>or on every connected uGASP clients (RMAIClientManager.ActorSessionID.BROADCAST_WITHOUT_GAME_LOGIC.getID())</li>
	 * <li>or on every connected uGASP clients and uGASP game logic (RMAIClientManager.ActorSessionID.BROADCAST_WITH_GAME_LOGIC.getID())</li>
	 * </ul>
	 * Methods that should be invoked must be implemented in the class RMAI client manager.
	 * Only instances methods can be invoked.
	 * Here are the types that can be used as parameters of this remote method:
	 * <ul>
	 * <li>Boolean</li>
	 * <li>Byte</li>
	 * <li>Short</li>
	 * <li>Integer</li>
	 * <li>Long</li>
	 * <li>Double</li>
	 * <li>Float</li>
	 * <li>String</li>
	 * <li>RMAISerializable (you should pay a special attention on recursion : never pass this object reference).</li>
	 * <li>Arrays of those several types</li>
	 * </ul>
	 * 
	 * @param recipientActorSessionID	the actorSession ID of the suitable recipient
	 * @param methodName				the name of the method that should be invoked
	 * @param param0					first parameter of the method that should be invoked
	 * @param param1					second parameter of the method that should be invoked
	 * @param param2					third parameter of the method that should be invoked
	 * @return boolean					true if the response code of the HTTP connection is OK
	 * 									false otherwise
	 */
	public void remoteMethodAsynchronousInvocation(
			short recipientClientActorSessionID, String methodName,
			Object param0, Object param1, Object param2) {
		if(dataTransmissionMgr.getUGASPClient().getID().isActorSessionIDValid()){
			// build the Object array
			Object[] params = {param0,param1,param2};
			// store an RMAI on the bufferOutputStream
			storeRMAI(recipientClientActorSessionID, methodName, params);
		}else{
			Log.w(getClass().getSimpleName(), "RMAI could not be invoked untill client has not retrieved a valid Actor Session ID");
		}
	}

	/**
	 * Invoke asynchronously a remote method without either :
	 * <ul>
	 * <li>on a uGASP Client, designated by its actor session id</li>
	 * <li>or on the uGASP game logic (RMAIClientManager.ActorSessionID.uGASP_GAME_LOGIC.getID())</li>
	 * <li>or on every connected uGASP clients (RMAIClientManager.ActorSessionID.BROADCAST_WITHOUT_GAME_LOGIC.getID())</li>
	 * <li>or on every connected uGASP clients and uGASP game logic (RMAIClientManager.ActorSessionID.BROADCAST_WITH_GAME_LOGIC.getID())</li>
	 * </ul>
	 * Methods that should be invoked must be implemented in the class RMAI client manager.
	 * Only instances methods can be invoked.
	 * Here are the types that can be used as parameters of this remote method:
	 * <ul>
	 * <li>Boolean</li>
	 * <li>Byte</li>
	 * <li>Short</li>
	 * <li>Integer</li>
	 * <li>Long</li>
	 * <li>Double</li>
	 * <li>Float</li>
	 * <li>String</li>
	 * <li>RMAISerializable (you should pay a special attention on recursion : never pass this object reference).</li>
	 * <li>Arrays of those several types</li>
	 * </ul>
	 * 
	 * @param recipientActorSessionID	the actorSession ID of the suitable recipient
	 * @param methodName				the name of the method that should be invoked
	 * @param param0					first parameter of the method that should be invoked
	 * @param param1					second parameter of the method that should be invoked
	 * @param param2					third parameter of the method that should be invoked
	 * @param param3					fourth parameter of the method that should be invoked
	 * @return boolean					true if the response code of the HTTP connection is OK
	 * 									false otherwise
	 */
	public void remoteMethodAsynchronousInvocation(
			short recipientClientActorSessionID, String methodName,
			Object param0, Object param1, Object param2, Object param3) {
		if(dataTransmissionMgr.getUGASPClient().getID().isActorSessionIDValid()){
			// build the Object array
			Object[] params = {param0,param1,param2,param3};
			// store an RMAI on the bufferOutputStream
			storeRMAI(recipientClientActorSessionID, methodName, params);
		}else{
			Log.w(getClass().getSimpleName(), "RMAI could not be invoked untill client has not retrieved a valid Actor Session ID");
		}

	}

	/**
	 * Invoke asynchronously a remote method either :
	 * <ul>
	 * <li>on a uGASP Client, designated by its actor session id</li>
	 * <li>or on the uGASP game logic (RMAIClientManager.ActorSessionID.uGASP_GAME_LOGIC.getID())</li>
	 * <li>or on every connected uGASP clients (RMAIClientManager.ActorSessionID.BROADCAST_WITHOUT_GAME_LOGIC.getID())</li>
	 * <li>or on every connected uGASP clients and uGASP game logic (RMAIClientManager.ActorSessionID.BROADCAST_WITH_GAME_LOGIC.getID())</li>
	 * </ul>
	 * Methods that should be invoked must be implemented in the class RMAI client manager.
	 * Only instances methods can be invoked.
	 * Here are the types that can be used as parameters of this remote method:
	 * <ul>
	 * <li>Boolean</li>
	 * <li>Byte</li>
	 * <li>Short</li>
	 * <li>Integer</li>
	 * <li>Long</li>
	 * <li>Double</li>
	 * <li>Float</li>
	 * <li>String</li>
	 * <li>RMAISerializable (you should pay a special attention on recursion : never pass this object reference).</li>
	 * <li>Arrays of those several types</li>
	 * </ul>
	 * 	
	 * @param recipientActorSessionID	the actorSession ID of the suitable recipient
	 * @param methodName				the name of the method that should be invoked
	 * @param param0					first parameter of the method that should be invoked
	 * @param param1					second parameter of the method that should be invoked
	 * @param param2					third parameter of the method that should be invoked
	 * @param param3					fourth parameter of the method that should be invoked
	 * @param param4					fifth parameter of the method that should be invoked
	 * @return boolean					true if the response code of the HTTP connection is OK
	 * 									false otherwise
	 */
	public void remoteMethodAsynchronousInvocation(
			short recipientClientActorSessionID, String methodName,
			Object param0, Object param1, Object param2, Object param3,
			Object param4) {
		if(dataTransmissionMgr.getUGASPClient().getID().isActorSessionIDValid()){
			// build the Object array
			Object[] params = {param0,param1,param2,param3,param4};
			// store an RMAI on the bufferOutputStream
			storeRMAI(recipientClientActorSessionID, methodName, params);	
		}else{
			Log.w(getClass().getSimpleName(), "RMAI could not be invoked untill client has not retrieved a valid Actor Session ID");
		}
	}

	/**
	 * Invoke asynchronously a remote method either :
	 * <ul>
	 * <li>on a uGASP Client, designated by its actor session id</li>
	 * <li>or on the uGASP game logic (RMAIClientManager.ActorSessionID.uGASP_GAME_LOGIC.getID())</li>
	 * <li>or on every connected uGASP clients (RMAIClientManager.ActorSessionID.BROADCAST_WITHOUT_GAME_LOGIC.getID())</li>
	 * <li>or on every connected uGASP clients and uGASP game logic (RMAIClientManager.ActorSessionID.BROADCAST_WITH_GAME_LOGIC.getID())</li>
	 * </ul>
	 * Methods that should be invoked must be implemented in the class RMAI client manager.
	 * Only instances methods can be invoked.
	 * Here are the types that can be used as parameters of this remote method:
	 * <ul>
	 * <li>Boolean</li>
	 * <li>Byte</li>
	 * <li>Short</li>
	 * <li>Integer</li>
	 * <li>Long</li>
	 * <li>Double</li>
	 * <li>Float</li>
	 * <li>String</li>
	 * <li>RMAISerializable (you should pay a special attention on recursion : never pass this object reference).</li>
	 * <li>Arrays of those several types</li>
	 * </ul>
	 * 
	 * @param recipientActorSessionID	the actorSession ID of the suitable recipient
	 * @param methodName				the name of the method that should be invoked
	 * @param param0					first parameter of the method that should be invoked
	 * @param param1					second parameter of the method that should be invoked
	 * @param param2					third parameter of the method that should be invoked
	 * @param param3					fourth parameter of the method that should be invoked
	 * @param param4					fifth parameter of the method that should be invoked
	 * @param param5					sixth parameter of the method that should be invoked
	 * @return boolean					true if the response code of the HTTP connection is OK
	 * 									false otherwise
	 */
	public void remoteMethodAsynchronousInvocation(
			short recipientClientActorSessionID, String methodName,
			Object param0, Object param1, Object param2, Object param3,
			Object param4, Object param5) {
		if(dataTransmissionMgr.getUGASPClient().getID().isActorSessionIDValid()){
			// build the Object array
			Object[] params = {param0,param1,param2,param3,param4,param5};
			// store an RMAI on the bufferOutputStream
			storeRMAI(recipientClientActorSessionID, methodName, params);
		}else{
			Log.w(getClass().getSimpleName(), "RMAI could not be invoked untill client has not retrieved a valid Actor Session ID");
		}	

	}

	/**
	 * Invoke asynchronously a remote method either :
	 * <ul>
	 * <li>on a uGASP Client, designated by its actor session id</li>
	 * <li>or on the uGASP game logic (RMAIClientManager.ActorSessionID.uGASP_GAME_LOGIC.getID())</li>
	 * <li>or on every connected uGASP clients (RMAIClientManager.ActorSessionID.BROADCAST_WITHOUT_GAME_LOGIC.getID())</li>
	 * <li>or on every connected uGASP clients and uGASP game logic (RMAIClientManager.ActorSessionID.BROADCAST_WITH_GAME_LOGIC.getID())</li>
	 * </ul>
	 * Methods that should be invoked must be implemented in the class RMAI client manager.
	 * Only instances methods can be invoked.
	 * Here are the types that can be used as parameters of this remote method:
	 * <ul>
	 * <li>Boolean</li>
	 * <li>Byte</li>
	 * <li>Short</li>
	 * <li>Integer</li>
	 * <li>Long</li>
	 * <li>Double</li>
	 * <li>Float</li>
	 * <li>String</li>
	 * <li>RMAISerializable (you should pay a special attention on recursion : never pass this object reference).</li>

	 * <li>Arrays of those several types</li>
	 * </ul>
	 * 
	 * @param recipientActorSessionID	the actorSession ID of the suitable recipient
	 * @param methodName				the name of the method that should be invoked
	 * @param param0					first parameter of the method that should be invoked
	 * @param param1					second parameter of the method that should be invoked
	 * @param param2					third parameter of the method that should be invoked
	 * @param param3					fourth parameter of the method that should be invoked
	 * @param param4					fifth parameter of the method that should be invoked
	 * @param param5					sixth parameter of the method that should be invoked
	 * @param param6					seventh parameter of the method that should be invoked
	 * @return boolean					true if the response code of the HTTP connection is OK
	 * 									false otherwise
	 */		
	public void remoteMethodAsynchronousInvocation(
			short recipientClientActorSessionID, String methodName,
			Object param0, Object param1, Object param2, Object param3,
			Object param4, Object param5, Object param6) {
		if(dataTransmissionMgr.getUGASPClient().getID().isActorSessionIDValid()){
			// build the Object array
			Object[] params = {param0,param1,param2,param3,param4,param5,param6};
			// store an RMAI on the bufferOutputStream
			storeRMAI(recipientClientActorSessionID, methodName, params);
		}else{
			Log.w(getClass().getSimpleName(), "RMAI could not be invoked untill client has not retrieved a valid Actor Session ID");
		}
	}
	
	
	/*
	 * This method is used by several remoteMethodAsynchronousInvocation to store a RMAI
	 * with the suitable parameters on the bufferOutputStream
	 */
	private void storeRMAI(short recipientClientActorSessionID, String methodName, Object[] params ){
		String invokedMethodXMLSerialized = null;
		try{
			// Serialization of "parameters" (method name + params, types and values) XMLRPC way
			invokedMethodXMLSerialized = rmaiUtils.methodInvocation(methodName, params);
		}catch(Exception e){
			e.printStackTrace();
		}
		if(dataTransmissionMgr.getStorageBufferOutputStream() != null){
				// if storage buffer is empty
				if(dataTransmissionMgr.getStorageBufferOutputStream().size() == 0){
					storeInsideEmptyBuffer(recipientClientActorSessionID, invokedMethodXMLSerialized);
				}else{
					storeInsideNonEmptyBuffer(recipientClientActorSessionID, invokedMethodXMLSerialized);
				}
		}else{
			Log.w(getClass().getSimpleName(), "Storage buffer is null, proceeding to reitinitialization of buffer.");
			// reinitialize storage buffer
			dataTransmissionMgr.reinitializeStorageBuffer();
		}	
	}
	
	
	/*
	 * In this method, we try to write on the storage buffer.
	 * If a problem occurs, it is simply flushed.
	 */
	private void storeInsideEmptyBuffer(short recipientClientActorSessionID,String invokedMethodXMLSerialized){
		// gets a lock on the storage buffer
		synchronized (dataTransmissionMgr.getStorageBufferOutputStream()) {
			try{
				dataTransmissionMgr.getStorageBufferOutputStream().writeByte(RMAI_REQUESTER_TYPE);
				dataTransmissionMgr.getStorageBufferOutputStream().writeShort(recipientClientActorSessionID);
				dataTransmissionMgr.getStorageBufferOutputStream().writeUTF(invokedMethodXMLSerialized);
			}catch(Throwable t1){
				Log.w(getClass().getSimpleName(), "Exception on storing of RMAI: it won't be sent to uGASP platform",t1);
				dataTransmissionMgr.resetStorageBuffer();
			}
		}
	}
	
	/*
	 * Method used to store RMAI in an intermediate buffer.
	 * If writing doesn't fail, then data are copied inside storage buffer.
	 */
	private void storeInsideNonEmptyBuffer(short recipientClientActorSessionID,String invokedMethodXMLSerialized){
		ByteArrayOutputStream baos 		= new ByteArrayOutputStream();
		DataOutputStream intermediate 	= new DataOutputStream(baos);
		boolean writeOK = true;
		try{
			intermediate.writeByte(RMAI_REQUESTER_TYPE);
			intermediate.writeShort(recipientClientActorSessionID);
			intermediate.writeUTF(invokedMethodXMLSerialized);
		}catch(Throwable t1){
			writeOK = false;
			Log.w(getClass().getSimpleName(), "Exception on storing of RMAI: it won't be sent to uGASP platform",t1);
		}
		// only if there wasn't any write exception
		if(writeOK){
			// gets a lock on the storage buffer
			synchronized (dataTransmissionMgr.getStorageBufferOutputStream()) {
				try {
					dataTransmissionMgr.getStorageBufferOutputStream().write(baos.toByteArray());
					// close buffers
					intermediate.close();
					baos.close();
				} catch(Throwable t){
					Log.w(getClass().getSimpleName(), "Exception on copying of RMAI: it won't be sent to uGASP platform",t);
				}
			}
		}
	}

}
