package org.playonline.network.datatransmission.protocol.rmai.tools;


import java.util.Map;


/**
 * Interface that should be implemented by objects designed for being called as parameter of RMAI.
 * Such objects also have to fulfill four other constraints:
 * <ul>
 * <li>fields that will be serialized must have a type handled by RMAI:
 * <ul>
 * <li>Boolean</li>
 * <li>Byte</li>
 * <li>Short</li>
 * <li>Integer</li>
 * <li>Long</li>
 * <li>Double</li>
 * <li>Float</li>
 * <li>String</li>
 * <li>RMAISerializable (you should pay a special attention on recursion : never pass this object reference).</li>
 * <li>Arrays of those several types</li>
 * </ul></li>
 * <li>they MUST have a no-argument constructor, since Java Reflection API is used to instantiate
 * those object on the parsing process</li>
 * <li>they have to be located into the org.ugasp.demonstrator.client.protocol.rmai.objects package, or in a sub package of this latter</li>
 * <li>location on the client and on the game logic side must be symmetric: an object located into the 
 * org.ugasp.demonstrator.client.protocol.rmai.objects package on the client side 
 * must be located into the org.ugasp.impl.games.demonstrator.protocol.rmai.objects on the game logic side.</li>
 * </ul>
 */
public interface RMAISerializable {
	
	/**
	 * Get a Map<String,Object> that should have an entry for each field to serialize:
	 * <ul>
	 * <li>key is a String representation of the name of the field</li>
	 * <li>value is an Object representation of the current value of this field.</li>
	 * </ul>
	 * Example:<br>
	 * An object which has the three following attributes:
	 * <p><blockquote><pre>
	 * Short 	id
	 * String 	name
	 * Integer 	position
	 * </pre></blockquote><p>
	 * should implement getSerializable() this way:
	 * <p><blockquote><pre>
	 * Map<String, Object> map = new HashMap<String, Object>();
	 * map.put("id",club);
	 * map.put("name",name);
	 * map.put("position",position);
	 * return map;
	 *</pre></blockquote><p>
	 *@return a map which contains names and values of fields that should be serialized
	 */
	public Map<String,Object> getSerializable();
	
	
	
	/**
	 * Initialize the instance with a Map<String,Object>, in order that:
	 * <ul>
	 * <li>keys of the map matches with names of fields</li>
	 * <li>values of the map matches with values that fields should be set to.</li>
	 * </ul>
	 * Example:<br>
	 * Lets continue with the same object used in getSerializable() documentation. Hence,
	 * initialize(Map<String,Object> map) should be implemented this way:
	 * <p><blockquote><pre>
	 * this.id		= (Byte)map.get("id");
	 * this.name	= (String)map.get("name");
	 * this.position	= (Integer)map.get("position");
	 * </pre></blockquote><p>
	 * 
	 * @param map	the map returned when calling getSerializable()
	 */
	public void initialize(Map<String,Object> map);
}
