package org.playonline.network.datatransmission.protocol.rmai.tools;


import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.StringWriter;
import java.util.ArrayList;

import org.xmlpull.v1.XmlPullParser;
import org.xmlpull.v1.XmlPullParserException;


/**
 * Class used to 
 * - serialize methods signatures as a Strings values of XML representations 
 * - create of InvokedMethod objects from those Strings. 
 */
public class RMAIUtils {
	
	private RMAICommon common;
	
	public RMAIUtils(){
		this.common = new RMAICommon();
	}
	
	/**
	 * 
	 * @param method
	 * @param params
	 * @return
	 * @throws Exception if a RMAISerializable object is attempted to be serialized while it is not located in the suitable package.
	 * @see RMAISerializable  
	 */
	public String methodInvocation(String method, Object[] params)throws Exception{
		StringWriter bodyWriter = new StringWriter();
			// get a lock on the XML serializer
			synchronized (common.getXmlSerializer()) {
				try {
					common.getXmlSerializer().setOutput(bodyWriter);
					common.getXmlSerializer().startDocument(null, null);
					common.getXmlSerializer().startTag(null, Tag.METHOD_CALL);
					// set method name
					common.getXmlSerializer().startTag(null, Tag.METHOD_NAME).text(method).endTag(null, Tag.METHOD_NAME);
					
					common.serializeParams(params);

					common.getXmlSerializer().endTag(null, Tag.METHOD_CALL);
					common.getXmlSerializer().endDocument();
				} catch (IllegalArgumentException e) {
					e.printStackTrace();
				} catch (IllegalStateException e) {
					e.printStackTrace();
				} catch (IOException e) {
					e.printStackTrace();
				}
				return bodyWriter.toString();
			}
	}
	
	/**
	 * Get an instance of InvokedMethod from a xml serialization stored in a String.
	 * This method parse the string to retrieve the method name as a String 
	 * and the parameters as an Object[].
	 * Method name and parameters are then used to return the matching InvokeMethod.
	 * 
	 * @param invokedMethodXMLSerialized	the String representing the method invoked, serialized xml way
	 * @return InvokedMethod				an InvokedMethod containing method name and parameters retrieved from the string.
	 * @throws	ClassNotFoundException 		when trying to deserialize an object which class that could be found
	 * @throws	InstantiationException 		when trying to deserialize an object which has no nullary constructor 	
	 * @throws	IllegalAccessException 		when trying to deserialize an object which nullary constructor is not public
	 */
	public InvokedMethod readMethodInvocation(String invokedMethodXMLSerialized) throws ClassNotFoundException,
																						InstantiationException,
																						IllegalAccessException{
		// get the pull parser matching with the xml serialized string
		XmlPullParser pullParser = xmlPullParserFromString(invokedMethodXMLSerialized);
		// begin the parsing of the xml pull parser
		try {
			// parsing method name
			pullParser.nextTag();
			pullParser.require(XmlPullParser.START_TAG, null, Tag.METHOD_CALL);
			pullParser.nextTag();
			pullParser.require(XmlPullParser.START_TAG, null, Tag.METHOD_NAME);

			String methodName = pullParser.nextText();
			//parsing parameters if needed
			pullParser.nextTag();
			// just for method with parameters
			if(pullParser.getName().equals(Tag.PARAMS)){
				pullParser.require(XmlPullParser.START_TAG, null, Tag.PARAMS);
				pullParser.nextTag(); // <param>
				
				ArrayList<Object> paramsList = new ArrayList<Object>();
				// parsing parameters (type & values)
				do {
					pullParser.require(XmlPullParser.START_TAG, null, Tag.PARAM);
					pullParser.nextTag(); // <value>

					Object param = common.getRMAISerializer().deserialize(pullParser);
					paramsList.add(param); // add to return value

					pullParser.nextTag();
					pullParser.require(XmlPullParser.END_TAG, null, Tag.PARAM);
					pullParser.nextTag(); // <param> or </params>
					
				} while (!pullParser.getName().equals(Tag.PARAMS)); // </params>
				return new InvokedMethod(methodName,paramsList);
			}else{
				// if method hasn't any parameters
				return new InvokedMethod(methodName,null);
			}
		} catch (XmlPullParserException e) {
			e.printStackTrace();
			return null;
		} catch (IOException e) {
			e.printStackTrace();
			return null;
		}
	}
	
	
	private XmlPullParser xmlPullParserFromString(String xmlSerializedString){
		// create an input stream matching with the string
		InputStream inputStream = new ByteArrayInputStream(xmlSerializedString.getBytes());
		// create a stream reader matching with th e input string
		Reader streamReader = new InputStreamReader(inputStream);
		// instantiate a xml parser
		XmlPullParser pullParser = new DefaultXMLParser();
		try {
			// affect the stream reader to the pull parser
			pullParser.setInput(streamReader);
			return pullParser;
		} catch (XmlPullParserException e) {
			e.printStackTrace();
			return null;
		}
	}
}
