package org.playonline.network.lobby;

import java.util.Vector;

import org.playonline.network.client.UGASPClient;

public interface LobbyMgr {

	UGASPClient 	getUGASPClient();
	boolean 		isInstanceQuit();
	void 			setInstanceQuit(boolean instanceQuit);
	
	/**
	 * This method should be called the first time a user connects to a specific application, 
	 * designated by its applicationID. It allows to specify a username and password for this application.
	 * An identifier called ActorID is created and saved into myActorID attribute, meaning that client 
	 * has properly registered itself in uGASP's database for this specific application.
	 * However, even if registration in the database is done, calling this method with the suitable parameters
	 * (username and password) allows to retrieve this actorID.
	 * 
	 *
	 * @param 	applicationID the ID of the application
	 * @param 	username the username wanted by the user
	 * @param 	password the password wanted by the user
	 * @return 	int 	the Actor ID, which is strictly positive if authentication informations provided are valid,
	 * 				0 otherwise.	
	 */
	int 			firstLogin(int applicationID, String username, String password);
	
	/**
	 * Log on uGASP Platform for a specific application.
	 * An identifier called sessionID is created and saved into mySessionID, which means that client is connected to uGASP,
	 * for a specific application. 
	 * To perform a login, you must have already called firstLogin, in order either to register in uGASP database and create your actorID, or
	 * just to retrieve your actorID if it is already created.
	 * 
	 * @param 	actorID the Actor ID which can be retrieve calling firstLogin()
	 * @param 	username the username chosen on the first login (and therefore registered in the database)
	 * @param 	password the password chosen on the first login (also registered in the database)
	 * 
	 * @return the Session ID of the user
	 */
	int 			login(int actorID, String username, String password);
	
	/**
	 * Returns a vector of ApplicationInstanceInfos, matching with application instances currently available on the uGASP platform.
	 * To join a specific instance:
	 * - get an (ApplicationInstanceInfos) element from the vector retrieved
	 * - get the applicationInstanceID of this element
	 * - use this applicationInstanceID as the second parameter of joinApplicationInstance(int sessionID, int applicationInstanceID)
	 * 
	 * @param sessionID the Session ID of the user
	 * 
	 * @return vector available applications instances, represented by ApplicationInstanceInfos
	 */
	Vector<ApplicationInstanceInfos>		getApplicationInstancesInfos(int sessionID);
	
	/**
	 * Join an application instance, designated by its actorSessionID.
	 * An identifier called actorSessionID is created and saved into myActorSessionID, which means that client is connected to 
	 * an application instance.
	 * 
	 * @param sessionID the Session ID of the user
	 * @param applicationInstanceID the ApplicationInstance ID of the instance which should be joined
	 * 
	 * @return the ActorSession ID of the user
	 */
	short 			joinApplicationInstance(int sessionID, int applicationInstanceID);
	
	/**
	 * Join randomly an open ApplicationInstance. 
	 * An identifier called actorSessionID is created and saved into myActorSessionID, 
	 * which means that client is connected to an application instance.
	 * 
	 * @param sID the Session ID of the user
	 * 
	 * @return the ActorSession ID of the user
	 */
	short 			joinRandomApplicationInstance(int sessionID);
	
	/**
	 * Create a public application instance, and connect to this instance.
	 * Instance created is designated by an applicationInstanceID. Thus, it can be retrieved calling
	 * getApplicationInstances (@see getApplicationInstances(int sessionID)).
	 * In addition to application instance creation, an identifier called actorSessionID is created and saved into myActorSessionID, 
	 * meaning that client is connected to the instance just created.
	 * 
	 * 
	 * @param sessionID the Session ID of the user
	 * @param minActors the minimum number of actors required to start, must be strictly positive
	 * @param maxActors the maximum number of actors
	 *  
	 * @return the ActorSession ID of the user
	 */
	short 			createApplicationInstance(int sessionID, int minActors, int maxActors);
	
	/**
	 * Create private application instance with the chosen actors, and connect to this instance.
	 * Instance created is designated by an applicationInstanceID. Thus, it can be retrieved calling
	 * getApplicationInstances (@see getApplicationInstances(int sessionID)).
	 * In addition to application instance creation, an identifier called actorSessionID is created and saved into myActorSessionID, 
	 * meaning that client is connected to the instance just created.
	 * 
	 * @param sID the Session ID of the user
	 * @param actors the table of ActorID representing the actors accepted in the game session
	 * @param minActors the minimum number of actors required to start, must be strictly positive
	 * @param maxActors the maximum number of actors
	 * 
	 * @return the ActorSession ID of the user
	 */
	short 			createPrivateApplicationInstance(int sessionID, String[] actors, int minActors, int maxActors);
	
	/**
	 * For application manual model only.
	 * Make the client which has created an application instance start it.
	 * 
	 * 
	 * @param 	actorSessionID	the actorSessionID of the client which has created the instance
	 * @return 	true if game has started properly, 
	 * 			false otherwise
	 */
	boolean 		startApplicationInstance(short actorSessionID);
	
	/**
	 * Make the client quit the application instance to which it is connected.
	 * WARNING
	 * -------
	 * If application model is set to MANUAL, making the owner call this method ENDS the application instance. 
	 * 
	 * @param aSID the ActorSession ID
	 */
	void 			quitApplicationInstance(short actorSessionID);
	
	/**
	 * WARNING
	 * -------
	 * 
	 * Should be called ONLY for gamelogic's life cycle set to PERMANENT.
	 * End the application instance.
	 * Set the value of instanceQuit to true in order to stop the data transmission routine.
	 * @param aSID
	 */
	void 			endPermanentApplicationInstance(short actorSessionID);
	
	/**
	 * Quit the platform.
	 * 
	 * @param sID the Session ID of the user
	 */
	void 			quitPlatform(int sessionID);
	
	/**
	 * While a user is connected to an application instance, this method allow it to change its pseudo.
	 * CAUTION: pseudo IS NOT the username used in firstLogin() and login() methods.
	 * 
	 * @param 	actorSessionID
	 * @param
	 * @return boolean
	 */
	String 			changePseudo(short actorSessionID, String pseudo);
	
}
