package org.playonline.network.lobby.http;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.HashMap;
import java.util.Map;
import java.util.Vector;

import org.playonline.network.client.UGASPClient;
import org.playonline.network.client.http.UGASPClientHttp;
import org.playonline.network.lobby.ApplicationInstanceInfos;
import org.playonline.network.lobby.LobbyMgr;
import org.playonline.network.lobby.ApplicationInstanceInfos.Player;
import org.playonline.network.tools.ConnectionTools;

import android.util.Log;

public class LobbyMgrHttp implements LobbyMgr {

	private static final int 	DEFAULT_USER_ID = 0;// ID used to match with OMA recommendations, this default value can be used for every clients
	private static final String APPLICATION_MANAGEMENT_SERVLET_PATH = "/Servlet";//servlet dedicated to application instances management (creation, start, end, join, ...)

	private static URL 			applicationInstanceManagementServletURL = getApplicationManagementInstanceServletURL();
	private HttpURLConnection 	applicationInstanceManagementConnection;
	
	private final UGASPClient	uGASPClient;
	private boolean 			instanceQuit;
	
	public LobbyMgrHttp(final UGASPClient uGASPClient){
		this.uGASPClient	= uGASPClient;
		this.instanceQuit 	= false;
	}
	
	private final static URL getApplicationManagementInstanceServletURL(){
		URL url = null;
		try {
			url = new URL(UGASPClientHttp.uGASPServletUrl+APPLICATION_MANAGEMENT_SERVLET_PATH);
		} catch (MalformedURLException e) {
			e.printStackTrace();
		}
		return url;
	}
	
	
	public static void updateUGASPUrl(){
		applicationInstanceManagementServletURL = getApplicationManagementInstanceServletURL();
	}
	
	
	public UGASPClient getUGASPClient() {
		return uGASPClient;
	}
	
	public boolean isInstanceQuit() {
		return instanceQuit;
	}
	
	public void setInstanceQuit(boolean instanceQuit) {
		this.instanceQuit = instanceQuit;
	}

	/**
	 * This method should be called the first time a user connects to a specific application, 
	 * designated by its applicationID. It allows to specify a username and password for this application.
	 * An identifier called ActorID is created and saved into myActorID attribute, meaning that client 
	 * has properly registered itself in uGASP's database for this specific application.
	 * However, even if registration in the database is done, calling this method with the suitable parameters
	 * (username and password) allows to retrieve this actorID.
	 * 
	 *
	 * @param 	applicationID the ID of the application
	 * @param 	username the username wanted by the user
	 * @param 	password the password wanted by the user
	 * @return 	int 	the Actor ID, which is strictly positive if authentication informations provided are valid,
	 * 				0 otherwise.	
	 */
	public int firstLogin(int applicationID, String username, String password) {
		String formData ="uid="+DEFAULT_USER_ID
		+"&appID="+applicationID
		+"&username="+username
		+"&password="+password;
		int actorID = -1;
		DataOutputStream 	applicationInstanceManagementOutputStream = null;
		DataInputStream 	applicationInstanceManagementInputStream = null;
		byte[] content = createContent("FirstLogin", formData);
		try {
			applicationInstanceManagementOutputStream = initApplicationInstanceManagementConnection(content);
			applicationInstanceManagementInputStream = applicationInstanceManagementServletRequest(applicationInstanceManagementOutputStream,content);
			if(applicationInstanceManagementInputStream != null){
				actorID = Integer.parseInt(applicationInstanceManagementInputStream.readShort()+"");
				uGASPClient.getID().setActorID(actorID);
			}else{
				Log.e(getClass().getSimpleName(), "Exception occured on the first login, cannot read response from uGASP Platform. Make sure it is properly launched.");
			}
		} catch (Throwable t) {
			Log.e(getClass().getSimpleName(), "Exception occured on the first login ", t);
			//e.printStackTrace();
		} finally { 
			// close data input stream 
			ConnectionTools.closeDataInputStream(applicationInstanceManagementInputStream);
			// close connection
			ConnectionTools.closeConnection(applicationInstanceManagementConnection);
		}
		return actorID;
	}



	/**
	 * Log on uGASP Platform for a specific application.
	 * An identifier called sessionID is created and saved into mySessionID, which means that client is connected to uGASP,
	 * for a specific application. 
	 * To perform a login, you must have already called firstLogin, in order either to register in uGASP database and create your actorID, or
	 * just to retrieve your actorID if it is already created.
	 * 
	 * @param 	actorID the Actor ID which can be retrieve calling firstLogin()
	 * @param 	username the username chosen on the first login (and therefore registered in the database)
	 * @param 	password the password chosen on the first login (also registered in the database)
	 * 
	 * @return the Session ID of the user
	 */
	public int login(int actorID, String username, String password) {
		String formData = "aID="+actorID
		+"&username="+username
		+"&password="+password;
		int sessionID = -1;
		DataOutputStream 	applicationInstanceManagementOutputStream = null;
		DataInputStream 	applicationInstanceManagementInputStream = null;
		byte[] content = createContent("Login", formData);
		try {
			applicationInstanceManagementOutputStream = initApplicationInstanceManagementConnection(content);
			applicationInstanceManagementInputStream = applicationInstanceManagementServletRequest(applicationInstanceManagementOutputStream,content);
			sessionID = Integer.parseInt(applicationInstanceManagementInputStream.readShort()+"");
			uGASPClient.getID().setSessionID(sessionID);
		} catch (Throwable t) {
			Log.e(this.getClass().getSimpleName(), "Login procedure failed", t);
		} finally { 
			// close data input stream 
			ConnectionTools.closeDataInputStream(applicationInstanceManagementInputStream);
			// close connection
			ConnectionTools.closeConnection(applicationInstanceManagementConnection);
		}
		return sessionID;
	}


	/**
	 * Returns a vector of ApplicationInstanceInfos, matching with application instances currently available on the uGASP platform.
	 * To join a specific instance:
	 * - get an (ApplicationInstanceInfos) element from the vector retrieved
	 * - get the applicationInstanceID of this element
	 * - use this applicationInstanceID as the second parameter of joinApplicationInstance(int sessionID, int applicationInstanceID)
	 * 
	 * @param sessionID the Session ID of the user
	 * 
	 * @return vector available applications instances, represented by ApplicationInstanceInfos
	 */
	public Vector<ApplicationInstanceInfos> getApplicationInstancesInfos(int sessionID) {
		String formData = "sID="+sessionID;
		Vector<ApplicationInstanceInfos> applicationInstancesInfos = null;
		DataOutputStream 	applicationInstanceManagementOutputStream = null;
		DataInputStream 	applicationInstanceManagementInputStream = null;
		byte[] content = createContent("Lobby", formData);
		try {
			applicationInstanceManagementOutputStream = initApplicationInstanceManagementConnection(content);
			applicationInstanceManagementInputStream = applicationInstanceManagementServletRequest(applicationInstanceManagementOutputStream,content);
			// then converted into a vector of ApplicationInstanceInfos
			applicationInstancesInfos = formLobbyDatas(applicationInstanceManagementInputStream);
		} catch (IOException e) {
			e.printStackTrace();
		} finally { 
			// close data input stream 
			ConnectionTools.closeDataInputStream(applicationInstanceManagementInputStream);
			// close connection
			ConnectionTools.closeConnection(applicationInstanceManagementConnection);
		}
		return applicationInstancesInfos;
	}


	/**
	 * Form the data contained on the DataInputStream of the lobby servlet response,
	 * returns a vector of ApplicationInstanceInfos objects.
	 * 
	 * @return the vector of ApplicationInstanceInfos objects
	 * 
	 * @throws IOException
	 */
	private Vector<ApplicationInstanceInfos> formLobbyDatas(DataInputStream dis) throws IOException {	    
		Vector<ApplicationInstanceInfos> appInstancesInfos = new Vector<ApplicationInstanceInfos>();

		// read the number of applications instances	
		short aIIDs = dis.readShort();

		while (aIIDs!= 0){
			// read the ApplicationInstance ID
			short aIID = dis.readShort();

			// read the minimum starting number of actors
			short min = dis.readShort();

			// read the number of actors in the application instance
			short playersNumber = dis.readShort(); 

			// read the maximum of actors
			short max = dis.readShort();

			// init the vector of actors
			Vector<Player> playerList = new Vector<Player>();

			while (playersNumber != 0){
				Player player = Player.getInstance(dis.readUTF(), dis.readInt());
				playerList.add(player);
				playersNumber--;
			}
			
			// read the number of optional infos
			short optionalInfosSize = dis.readShort();
			
			Map<String, String> optionalInfos = null;
			if(optionalInfosSize != 0){
				optionalInfos = new HashMap<String, String>();
				for (int i = 0; i < optionalInfosSize; i++) {
					String key = dis.readUTF();
					String value = dis.readUTF();
					optionalInfos.put(key, value);
				}
			}

			// form the above datas
			ApplicationInstanceInfos appInfos = ApplicationInstanceInfos.getInstance(
					Integer.parseInt(aIID+""),
					min,
					max,
					playerList.size(),
					playerList,
					optionalInfos);

			appInstancesInfos.addElement(appInfos);
			aIIDs--;
		}

		// return the vector of ApplicationInstanceInfos objects
		return appInstancesInfos;
	}


	/**
	 * Join an application instance, designated by its actorSessionID.
	 * An identifier called actorSessionID is created and saved into myActorSessionID, which means that client is connected to 
	 * an application instance.
	 * 
	 * @param sessionID the Session ID of the user
	 * @param applicationInstanceID the ApplicationInstance ID of the instance which should be joined
	 * 
	 * @return the ActorSession ID of the user
	 */
	public short joinApplicationInstance(int sessionID, int applicationInstanceID) {
		String formData = "sID="+sessionID+"&aIID="+applicationInstanceID;
		short actorSessionID = -1; 
		DataOutputStream 	applicationInstanceManagementOutputStream = null;
		DataInputStream 	applicationInstanceManagementInputStream = null;
		byte[] content = createContent("Join", formData);
		try {
			applicationInstanceManagementOutputStream = initApplicationInstanceManagementConnection(content);
			applicationInstanceManagementInputStream = applicationInstanceManagementServletRequest(applicationInstanceManagementOutputStream,content);
			actorSessionID = applicationInstanceManagementInputStream.readShort();
			uGASPClient.getID().setActorSessionID(actorSessionID);				
		} catch (IOException e) {
			e.printStackTrace();
		} finally {
			// close data input stream 
			ConnectionTools.closeDataInputStream(applicationInstanceManagementInputStream);
			// close connection
			ConnectionTools.closeConnection(applicationInstanceManagementConnection);
		}
		return actorSessionID;
	}


	/**
	 * Join randomly an open ApplicationInstance. 
	 * An identifier called actorSessionID is created and saved into myActorSessionID, 
	 * which means that client is connected to an application instance.
	 * 
	 * @param sID the Session ID of the user
	 * 
	 * @return the ActorSession ID of the user
	 */
	public short joinRandomApplicationInstance(int sessionID) {
		String formData = "sID="+sessionID+"&aIID=";
		short actorSessionID = -1;
		DataOutputStream 	applicationInstanceManagementOutputStream = null;
		DataInputStream 	applicationInstanceManagementInputStream = null;
		byte[] content = createContent("Join", formData);
		try {
			applicationInstanceManagementOutputStream = initApplicationInstanceManagementConnection(content);
			applicationInstanceManagementInputStream = applicationInstanceManagementServletRequest(applicationInstanceManagementOutputStream,content);
			actorSessionID = applicationInstanceManagementInputStream.readShort();
			uGASPClient.getID().setActorSessionID(actorSessionID);
		} catch (IOException e) {
			e.printStackTrace();
		} finally {
			// close data input stream 
			ConnectionTools.closeDataInputStream(applicationInstanceManagementInputStream);
			// close connection
			ConnectionTools.closeConnection(applicationInstanceManagementConnection);
		}
		return actorSessionID;
	}


	/**
	 * Create a public application instance, and connect to this instance.
	 * Instance created is designated by an applicationInstanceID. Thus, it can be retrieved calling
	 * getApplicationInstances (@see getApplicationInstances(int sessionID)).
	 * In addition to application instance creation, an identifier called actorSessionID is created and saved into myActorSessionID, 
	 * meaning that client is connected to the instance just created.
	 * 
	 * 
	 * @param sessionID the Session ID of the user
	 * @param minActors the minimum number of actors required to start, must be strictly positive
	 * @param maxActors the maximum number of actors
	 *  
	 * @return the ActorSession ID of the user
	 */
	public short createApplicationInstance(int sessionID, int minActors, int maxActors) {
		String formData = "sID="+sessionID
		+"&minActors="+minActors
		+"&maxActors="+maxActors
		+"&actors=";
		short actorSessionID = -1;
		DataOutputStream 	applicationInstanceManagementOutputStream = null;
		DataInputStream 	applicationInstanceManagementInputStream = null;
		byte[] content = createContent("Create", formData);
		try {
			applicationInstanceManagementOutputStream = initApplicationInstanceManagementConnection(content);
			applicationInstanceManagementInputStream = applicationInstanceManagementServletRequest(applicationInstanceManagementOutputStream,content);
			actorSessionID= applicationInstanceManagementInputStream.readShort();
			uGASPClient.getID().setActorSessionID(actorSessionID);
			// Set back the value of instanceQuit if needed
			if(instanceQuit){
				instanceQuit = false;
			}
			return uGASPClient.getID().getActorSessionID();
		} catch (IOException e) {
			e.printStackTrace();
		} finally { 
			// close data input stream 
			ConnectionTools.closeDataInputStream(applicationInstanceManagementInputStream);
			// close connection
			ConnectionTools.closeConnection(applicationInstanceManagementConnection);
		}
		return actorSessionID;
	}


	/**
	 * Create private application instance with the chosen actors, and connect to this instance.
	 * Instance created is designated by an applicationInstanceID. Thus, it can be retrieved calling
	 * getApplicationInstances (@see getApplicationInstances(int sessionID)).
	 * In addition to application instance creation, an identifier called actorSessionID is created and saved into myActorSessionID, 
	 * meaning that client is connected to the instance just created.
	 * 
	 * @param sID the Session ID of the user
	 * @param actors the table of ActorID representing the actors accepted in the game session
	 * @param minActors the minimum number of actors required to start, must be strictly positive
	 * @param maxActors the maximum number of actors
	 * 
	 * @return the ActorSession ID of the user
	 */
	public short createPrivateApplicationInstance(int sessionID, String[] actors, int minActors, int maxActors) {
		String actorsHttp ="";
		for(int i=0; i<actors.length; i++){
			actorsHttp = actorsHttp + "_"+ actors[i];
		}
		String formData = "sID="+sessionID
		+"&minActors="+minActors
		+"&maxActors="+maxActors
		+"&actors="+actorsHttp;
		short actorSessionID = -1;
		DataOutputStream 	applicationInstanceManagementOutputStream = null;
		DataInputStream 	applicationInstanceManagementInputStream = null;
		byte[] content = createContent("Create", formData);
		try {
			applicationInstanceManagementOutputStream = initApplicationInstanceManagementConnection(content);
			applicationInstanceManagementInputStream = applicationInstanceManagementServletRequest(applicationInstanceManagementOutputStream,content);
			actorSessionID = applicationInstanceManagementInputStream.readShort();
			uGASPClient.getID().setActorSessionID(actorSessionID);
			// Set back the value of instanceQuit if needed
			if(instanceQuit){
				instanceQuit = false;
			}
		} 
		catch (Exception e) {
			e.printStackTrace();
		}
		finally {
			// close data input stream 
			ConnectionTools.closeDataInputStream(applicationInstanceManagementInputStream);
			// close connection
			ConnectionTools.closeConnection(applicationInstanceManagementConnection);
		}
		return actorSessionID;
	}


	/**
	 * While a user is connected to an application instance, this method allow it to change its pseudo.
	 * CAUTION: pseudo IS NOT the username used in firstLogin() and login() methods.
	 * 
	 * @param 	actorSessionID
	 * @param
	 * @return boolean
	 */
	public String changePseudo(short actorSessionID, String pseudo){
		String formData = "aSID="+actorSessionID+"&name="+pseudo;
		String newPseudo = null;
		DataOutputStream 	applicationInstanceManagementOutputStream = null;
		DataInputStream 	applicationInstanceManagementInputStream = null;
		byte[] content = createContent("Name", formData);
		try {
			applicationInstanceManagementOutputStream = initApplicationInstanceManagementConnection(content);
			applicationInstanceManagementInputStream = applicationInstanceManagementServletRequest(applicationInstanceManagementOutputStream,content);
			newPseudo = applicationInstanceManagementInputStream.readUTF();
		} catch (IOException e) {
			e.printStackTrace();
		}
		finally {
			// close data input stream 
			ConnectionTools.closeDataInputStream(applicationInstanceManagementInputStream);
			// close connection
			ConnectionTools.closeConnection(applicationInstanceManagementConnection);
		}
		return newPseudo;
	}


	/**
	 * For application manual model only.
	 * Make the client which has created an application instance start it.
	 * 
	 * 
	 * @param 	actorSessionID	the actorSessionID of the client which has created the instance
	 * @return 	true if game has started properly, 
	 * 			false otherwise
	 */
	public boolean startApplicationInstance(short actorSessionID){
		String formData = "aSID="+actorSessionID;
		boolean isCreated = false;
		DataOutputStream 	applicationInstanceManagementOutputStream = null;
		DataInputStream 	applicationInstanceManagementInputStream = null;
		byte[] content = createContent("StartAI", formData);
		try {
			applicationInstanceManagementOutputStream = initApplicationInstanceManagementConnection(content);
			applicationInstanceManagementInputStream = applicationInstanceManagementServletRequest(applicationInstanceManagementOutputStream,content);
			isCreated = applicationInstanceManagementInputStream.readBoolean();
		} catch (IOException e) {
			e.printStackTrace();
		} finally {
			// close data input stream 
			ConnectionTools.closeDataInputStream(applicationInstanceManagementInputStream);
			// close connection
			ConnectionTools.closeConnection(applicationInstanceManagementConnection);
		}
		return isCreated;
	}


	/**
	 * Make the client quit the application instance to which it is connected.
	 * WARNING
	 * -------
	 * If application model is set to MANUAL, making the owner call this method ENDS the application instance. 
	 * 
	 * @param aSID the ActorSession ID
	 */
	public void quitApplicationInstance(short actorSessionID){
		String formData = "aSID="+actorSessionID;
		DataOutputStream 	applicationInstanceManagementOutputStream = null;
		byte[] content = createContent("QuitAI", formData);
		try {
			applicationInstanceManagementOutputStream = initApplicationInstanceManagementConnection(content);
			applicationInstanceManagementServletRequest(applicationInstanceManagementOutputStream,content);
			// to stop the get event routine of inherited classes
			instanceQuit = true;
			Log.i(this.getClass().getSimpleName(), "Application instance quit");
		} catch (IOException e) {
			e.printStackTrace();
		} finally{
			// close connection
			ConnectionTools.closeConnection(applicationInstanceManagementConnection);
		}
	}


	/**
	 * WARNING
	 * -------
	 * 
	 * Should be called ONLY for gamelogic's life cycle set to PERMANENT.
	 * End the application instance.
	 * Set the value of instanceQuit to true in order to stop the data transmission routine.
	 * @param aSID
	 */
	public void endPermanentApplicationInstance(short actorSessionID){
		String formData = "aSID="+actorSessionID;
		DataOutputStream 	applicationInstanceManagementOutputStream = null;
		byte[] content = createContent("EndAI", formData);
		try {
			applicationInstanceManagementOutputStream = initApplicationInstanceManagementConnection(content);
			applicationInstanceManagementServletRequest(applicationInstanceManagementOutputStream,content);
			// to stop the get event routine of inherited classes
			instanceQuit = true;
			Log.i(this.getClass().getSimpleName(), "Permanent Application Instance ended");
		} catch (IOException e) {
			e.printStackTrace();
		} finally{
			// close connection
			ConnectionTools.closeConnection(applicationInstanceManagementConnection);
		}
	}


	/**
	 * Quit the platform.
	 * 
	 * @param sID the Session ID of the user
	 */
	public void quitPlatform(int sessionID) {
		String formData = "sID="+sessionID;
		DataOutputStream 	applicationInstanceManagementOutputStream = null;
		byte[] content = createContent("Quit", formData);
		try {
			applicationInstanceManagementOutputStream = initApplicationInstanceManagementConnection(content);
			applicationInstanceManagementServletRequest(applicationInstanceManagementOutputStream,content);
			// to stop the get event routine of inherited classes
			instanceQuit = true;
			Log.i(this.getClass().getSimpleName(), "uGASP Platform quit");
		} catch (IOException e) {
			e.printStackTrace();
		} finally{
			// close connection
			ConnectionTools.closeConnection(applicationInstanceManagementConnection);
		}
	}
	
	/*
	 * As we cannot synchronized on connection since they it is not instantiate,
	 * this method is declared synchronized.
	 * This method initiates the applicationInstanceManagementConnection.
	 * 
	 * @param	content the content of the request 
	 * @return 	the DataOutputStream that shall be used to write the request
	 * @throws 	IOException
	 */
	private synchronized DataOutputStream initApplicationInstanceManagementConnection(byte[] content) throws IOException{
		DataOutputStream dos = null;
		applicationInstanceManagementConnection = (HttpURLConnection) applicationInstanceManagementServletURL.openConnection();
		// setting connection
		applicationInstanceManagementConnection.setDoOutput(true);
		applicationInstanceManagementConnection.setDoInput(true);
		applicationInstanceManagementConnection.setRequestMethod("POST");
		applicationInstanceManagementConnection.setRequestProperty("Content-Type", "application/x-www-form-urlencoded");
		applicationInstanceManagementConnection.setRequestProperty("Accept", "application/octet-stream");
		// TODO Should be confirm with tests but it seems "Content-Length" shouldn't be set in ANDROID environment
		//applicationInstanceManagementConnection.setRequestProperty("Content-Length",Integer.toString(content.length));
		// get DataOutputStream from the connection
		dos = new DataOutputStream(applicationInstanceManagementConnection.getOutputStream());
		return dos;
	}
	
	
	/*
	 * Send a request to the servlet dedicated to application instances management, with the http post parameters.
	 * Return the servlet response DataInputStream.
	 * Note that connection MUST be already initialized with initApplicationInstanceManagementConnection(). 
	 * 
	 * @param	content the content of the request 
	 * @return 	the DataInputStream containing response from the platform
	 * @throws 	IOException
	 */
	private DataInputStream applicationInstanceManagementServletRequest(DataOutputStream applicationInstanceManagementOutputStream, byte[] content) throws IOException{
		DataInputStream dis= null;
		// get a lock on the application instance management connection, which MUST be already initialized
		synchronized (applicationInstanceManagementConnection) {
			applicationInstanceManagementOutputStream.write(content);
			// close data output stream --> raise an error on the simulator... but seem to speed the treatment on the device !
			//ConnectionTools.closeDataOutputStream(applicationInstanceManagementOutputStream);
			// return connection's data input stream if response code is OK 
			if (applicationInstanceManagementConnection.getResponseCode() == HttpURLConnection.HTTP_OK){
				dis = new DataInputStream(applicationInstanceManagementConnection.getInputStream());
			}else{
				Log.w(getClass().getSimpleName(), "Invalid response from uGASP, code = "+applicationInstanceManagementConnection.getResponseCode());
			}
		}
		return dis;
	}

	
	/*
	 * Create the content of a request from a request type (FirstLogin, Login...) and its parameters.
	 * 
	 * @param requestType	the type of the request
	 * @param formData		the parameters of the request
	 * @return				a byte[] representing the content of the request.
	 */
	private static byte[] createContent(String requestType, String formData){
		formData="from=2&mode="+requestType+"&"+formData;
		return formData.getBytes();
	}
}
