package org.playonline.ui.basic.interaction;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.concurrent.ConcurrentHashMap;

import org.playonline.network.client.RMAIClientManager;
import org.playonline.ui.basic.map.PlayOnlineMapActivity;
import org.playonline.ui.basic.map.handler.MapViewMessage;
import org.playonline.ui.basic.profile.Character;

import android.util.Log;

import com.google.android.maps.GeoPoint;

public class CharactersManager {

	private final 	PlayOnlineMapActivity 		mapActivity;
	private 		GeoPoint 					myCharaterLocation;
	private  		Map<Short,Character>  		characterByActorSessionId;
	private 		Map<GeoPoint,List<Short>> 	charactersListOnLocation;

	private CharactersManager(PlayOnlineMapActivity mapActivity) {
		this.mapActivity 				= mapActivity;
		this.characterByActorSessionId 	= new ConcurrentHashMap<Short,Character>();
		this.charactersListOnLocation 	= new ConcurrentHashMap<GeoPoint,List<Short>>();
		this.myCharaterLocation			= mapActivity.getOverlayManager().getCharacterOverlay().getLocation();
	}


	public static CharactersManager getInstance(PlayOnlineMapActivity mapActivity){
		return new CharactersManager(mapActivity);
	}
	
	
	public String getNameByActorSessionId(Short actorSessionId){
		String name = null;
		Character character = characterByActorSessionId.get(actorSessionId);
		if(character != null){
			name = character.getName();
		}
		return name;
	}
	
	
	public Short getActorSessionIdByName(String name){
		Short actorSessionId = null;
		for (Entry<Short,Character> entry : characterByActorSessionId.entrySet()){
			if(entry.getValue().getName().equals(name)){
				actorSessionId = entry.getKey();
				break;
			}
		}
		return actorSessionId;
	}
	
	
	public boolean isMyCharacterLeader(){
		boolean isLeader = false;
		for (Entry<Short,Character> character : characterByActorSessionId.entrySet()) {
			if(character.getValue().getLeaderActorSessionID() == (mapActivity.getUGASPClient().getID().getActorSessionID())){
				isLeader = true;
				break;
			}
		}
		return isLeader;
	}
	
	
	public void removedCharacter(Short actorSessionId){
		// if character was my leader
		if(mapActivity.getCharacter().isAttachedToLeader() && mapActivity.getCharacter().getLeaderActorSessionID() == actorSessionId.shortValue()){
			mapActivity.getCharacter().detachFromLeader();
		}
		characterByActorSessionId.remove(actorSessionId);
		removeCharacterFromLocation(actorSessionId);
		// ask UI thread to update overlays
		mapActivity.getMapViewHandler().sendMessage(MapViewMessage.updateCharactersOverlays());
	}
	
	
	public void putCharacter(Short actorSessionId, String name){
		Character character = Character.getInstance(name);
		characterByActorSessionId.put(actorSessionId, character);
	}
	
	
	public void moveCharacterToLocation(GeoPoint location, Short characterActorSessionId){
		// we test if characters are following the one designed by characterActorSessionId
		for (Entry<Short,Character> entry : characterByActorSessionId.entrySet()){
			// if its the case, their location are updated
			if(entry.getValue().isAttachedToLeader() && entry.getValue().getLeaderActorSessionID() == characterActorSessionId.shortValue()){
				putCharacterOnLocation(location,entry.getKey());
			}
		}
		// character location is updated too
		putCharacterOnLocation(location, characterActorSessionId);
		// if character is my leader, my position is updated too
		if(mapActivity.getCharacter().isAttachedToLeader() && mapActivity.getCharacter().getLeaderActorSessionID() == characterActorSessionId.shortValue()){
			setMyCharacterLocation(location);
		}else{
			// ask UI thread to update overlays
			mapActivity.getMapViewHandler().sendMessage(MapViewMessage.updateCharactersOverlays());
		}
	}
	
	
	private void putCharacterOnLocation(GeoPoint location, Short characterActorSessionId){
		// former character's location is removed
		removeCharacterFromLocation(characterActorSessionId);
		// and replaced by its new location
		if(charactersListOnLocation.get(location) == null){
			List<Short> charactersList = new ArrayList<Short>();
			charactersList.add(characterActorSessionId);
			charactersListOnLocation.put(location, charactersList);
		}else{
			List<Short> charactersList = charactersListOnLocation.get(location);
			if(!charactersList.contains(characterActorSessionId)){
				charactersList.add(characterActorSessionId);
			}
		}
	}
	
	public void removeCharacterFromLocation(Short characterActorSessionId){
		GeoPoint characterLocation = getCharacterLocation(characterActorSessionId);
		if(characterLocation != null){
			List<Short> charactersList = charactersListOnLocation.get(characterLocation);
			charactersList.remove(characterActorSessionId);
			// if there are no more characters on that location
			if(charactersList.size() == 0){
				charactersListOnLocation.remove(characterLocation);
			}
		}
	}
	
	private GeoPoint getCharacterLocation(Short characterActorSessionId){
		GeoPoint characterLocation = null;
		for (Entry<GeoPoint,List<Short>> charactersList : charactersListOnLocation.entrySet()) {
			if(charactersList.getValue().contains(characterActorSessionId)){
				characterLocation = charactersList.getKey();
				break;
			}
		}
		return characterLocation;
	}
	
	
	public GeoPoint getMyCharacterLocation(){
		return myCharaterLocation;
	}
	
	
	public void setMyCharacterLocation(GeoPoint newLocation){
		// check if my character has members, and if its the case update their location
		putMyMembersOnLocation(newLocation);
		this.myCharaterLocation = newLocation;
		// if player is connected to uGASP but not attached to a leader, its location is sent to platform
		if(mapActivity.getUGASPClient() != null && !mapActivity.getCharacter().isAttachedToLeader()){
			mapActivity.getUGASPClient().getDataTransmissionMgr().getRMAIMgr().remoteMethodAsynchronousInvocation(
					RMAIClientManager.ActorSessionID.uGASP_GAME_LOGIC.getID(), 
					"updateCharacterLocation",
					new Integer(newLocation.getLatitudeE6()),
					new Integer(newLocation.getLongitudeE6()));
		}
		// if the player is attached to a leader, its the leader that called this method
		if(mapActivity.getCharacter().isAttachedToLeader()){
			// so with have to animate to its new location
			mapActivity.getMapViewHandler().sendMessage(MapViewMessage.animateToMyLocationMessage());
		}
		// ask UI thread to update overlays
		mapActivity.getMapViewHandler().sendMessage(MapViewMessage.updateCharactersOverlays());
	}
	
	
	private void putMyMembersOnLocation(GeoPoint newLocation){
		List<Short> charactersOnLocationList = getCharactersListInLocation(this.myCharaterLocation);
		if(charactersOnLocationList != null){
			List<Short> copyOfCharactersOnLocationList = new ArrayList<Short>();
			copyOfCharactersOnLocationList.addAll(charactersOnLocationList);
			for (Short actorSessionId : copyOfCharactersOnLocationList) {
				Character character = getCharacterByActorSessionId(actorSessionId);
				// if my character is the leader of this character
				if(character != null && character.isAttachedToLeader() && character.getLeaderActorSessionID() == mapActivity.getUGASPClient().getID().getActorSessionID()){
					// character position is updated
					putCharacterOnLocation(newLocation,actorSessionId);
				}
			}
		}
	}
	
	
	public Character getCharacterByActorSessionId(Short actorSessionId){
		return characterByActorSessionId.get(actorSessionId);
	}
	
	
	public List<Short> getCharactersListInLocation(GeoPoint location){
		return charactersListOnLocation.get(location);
	}
	
	
	public Map<GeoPoint,List<Short>> getCharactersListInLocationMap(){
		return charactersListOnLocation;
	}
	
	
	public void setCharactersListInLocationMap(Map<GeoPoint,List<Short>> map){
		this.charactersListOnLocation = map;
		displayCharactersListInLocationMap();
		// ask UI thread to update overlays
		mapActivity.getMapViewHandler().sendMessage(MapViewMessage.updateCharactersOverlays());
	}
	
	
	public void setCharacterByActorSessionIdMap(Map<Short,Character> map){
		this.characterByActorSessionId = map;
		displayCharacterByActorSessionIdMap();
	}
	
	
	private void displayCharactersListInLocationMap(){
		Log.i(getClass().getSimpleName(), "------    charactersListOnLocation size = "+charactersListOnLocation.size());
		synchronized (charactersListOnLocation) {
			for(Entry<GeoPoint, List<Short>> entry : charactersListOnLocation.entrySet()){
				Log.i(getClass().getSimpleName(), "------    Location = "+entry.getKey()+" --- Characters = "+entry.getValue());
			}
		}
	}
	
	
	private void displayCharacterByActorSessionIdMap(){
		Log.i(getClass().getSimpleName(), "------    characterByActorSessionId size = "+characterByActorSessionId.size());
		synchronized (characterByActorSessionId) {
			for(Entry<Short,Character> entry : characterByActorSessionId.entrySet()){
				Log.i(getClass().getSimpleName(), "------    ActorSessionId = "+entry.getKey()+" --- Character = "+entry.getValue());
			}
		}
	}
	

}
