package org.playonline.ui.basic.map.overlay;

import java.util.ArrayList;
import java.util.List;

import org.playonline.ui.basic.R;
import org.playonline.ui.basic.map.PlayOnlineMapActivity;
import org.playonline.ui.basic.map.handler.MapViewMessage;

import android.location.Location;
import android.os.Message;
import android.util.Log;

import com.google.android.maps.GeoPoint;
import com.google.android.maps.ItemizedOverlay;
import com.google.android.maps.OverlayItem;


public class HiddenElementOverlay extends ItemizedOverlay<OverlayItem>{

	private static int MARKER = R.drawable.question_mark;
	private static float DEFAULT_DISTANCE_TO_UNBLOCK = 700;
	
	public static final int ADD_OVERLAY_WHAT = 1;
	public static final int REMOVE_OVERLAY_WHAT = 2;
	
	private final static List<HiddenElementOverlay> hiddenElementsList = initHiddenElementsList();
	
	private final ArrayList<OverlayItem> mOverlays;
	private final GeoPoint location;
	private final OverlayItem overlayItem;
	private final String name,description;
	private final float distanceToUnblock;
	private final PlayOnlineMapActivity mapActivity;
	
	
	public HiddenElementOverlay(PlayOnlineMapActivity mapActivity, GeoPoint location, String name, String description) {
		super(boundCenterBottom(mapActivity.getResources().getDrawable(MARKER)));
		this.mapActivity = mapActivity;
		this.name = name;
		this.description = description;
		mOverlays = new ArrayList<OverlayItem>();
		this.location = location;
		this.distanceToUnblock = DEFAULT_DISTANCE_TO_UNBLOCK;
		overlayItem = new OverlayItem(location, name, "");
		this.addOverlay(overlayItem);
	}

	
	/**
	 * Check whether there are hidden element to unblock.
	 * If there are, a message is sent to the UI thread in order that it add the matching overlay to the map view.
	 * @param mapActivity
	 */
	public static void unblockElement(PlayOnlineMapActivity mapActivity){
		GeoPoint location = mapActivity.getCharactersManager().getMyCharacterLocation();
		double startLatitude, startLongitude, endLatitude, endLongitude;
		float[] distance = new float[1];
		startLatitude 	= location.getLatitudeE6()/1E6;
		startLongitude 	= location.getLongitudeE6()/1E6;
		for (HiddenElementOverlay overlay : hiddenElementsList) {
			endLatitude 		= overlay.getLocation().getLatitudeE6()/1E6;
			endLongitude		= overlay.getLocation().getLongitudeE6()/1E6;
			// computes distance between character overlay and hidden element
			Location.distanceBetween(startLatitude, startLongitude, endLatitude, endLongitude, distance);
			Log.i(HiddenElementOverlay.class.getSimpleName(), "+++++ Distance to hidden element = "+distance[0]+" meters");
			if(distance[0]<=overlay.distanceToUnblock){
				// get a message containing the hashCode of the overlay to the map activity, in order to add it to its overlays' list
				// in order to remove it from its overlays' list
				Message msg = MapViewMessage.addHiddenElementOverlayMessage(overlay.hashCode());
				// send message to the mapActivity
				mapActivity.getMapViewHandler().sendMessage(msg);
			}
		}
	}
	
	/**
	 * Return the HiddenElementOverlay matching with the given hashCode, and remove it from the list of hidden elements.
	 * 
	 * @param elementHashCode
	 * @return
	 */
	public static HiddenElementOverlay getElementFromHashCode(int elementHashCode){
		HiddenElementOverlay result = null;
		for (HiddenElementOverlay overlay : hiddenElementsList) {
			if(overlay.hashCode() == elementHashCode){
				// retrieve overlay
				result = overlay;
				break;
			}
		}
		// removed from the hidden elements list
		if(result != null){
			hiddenElementsList.remove(result);
		}
		return result;
	}
	
	
	private static List<HiddenElementOverlay> initHiddenElementsList() {
		return new ArrayList<HiddenElementOverlay>();
	}


	public void addOverlay(OverlayItem overlay) {
		mOverlays.add(overlay);
		populate();
	}

	@Override
	protected OverlayItem createItem(int i) {
		return mOverlays.get(i);
	}

	@Override
	public int size() {
		return mOverlays.size();
	}

	@Override
	protected boolean onTap(int index) {
		// if character is not attached to a leader
		if(!mapActivity.getCharacter().isAttachedToLeader()){
			mapActivity.getNewElementsAlertDialog().displayNewElementInInventory(name,description, MARKER);
			// get a message containing the hashCode of the current HiddenElementOverlay to the map activity, 
			// in order to remove it from its overlays' list
			Message msg = MapViewMessage.removeHiddenElementOverlayMessage(this.hashCode());
			// send message to the mapActivity
			mapActivity.getMapViewHandler().sendMessage(msg);
			if(mapActivity.getCharactersManager().isMyCharacterLeader()){
				// send RMAI to group members to signal that element is caught
				mapActivity.getGroupManager().makeGroupCatchHiddenElement(name, description, MARKER, this.hashCode());
			}
		}
		
		return true;
	}
	
	
	public GeoPoint getLocation(){
		return location;
	}


	public float getDistanceToUnblock() {
		return distanceToUnblock;
	}
	
	
	public static void addHiddentElementOverlay(HiddenElementOverlay overlay){
		hiddenElementsList.add(overlay);
	}
	
	
	public boolean equals(Object o){
		if(!(o instanceof HiddenElementOverlay))
			return false;
		HiddenElementOverlay overlay = (HiddenElementOverlay)o;
		return (overlay.location.getLatitudeE6() == this.location.getLatitudeE6()) 
				&& (overlay.location.getLongitudeE6() == this.location.getLongitudeE6())
				&& (overlay.distanceToUnblock == this.distanceToUnblock)
				&& (overlay.name.equals(this.name));
	}

	
	public int hashCode(){
		// Values 17 and 37 are from recommendation 8
		// in the book "Effective Java - programming language guide" by Joshua Block.
		int result = 17;
		result = 37 * result + location.getLatitudeE6();
		result = 37 * result + location.getLongitudeE6();
		result = 37 * result + (int)distanceToUnblock;
		result = 37 * result + name.hashCode();
		result = 37 * result + description.hashCode();
		return result;
	}
}