package org.playonline.ui.basic.map.overlay;

import java.util.ArrayList;
import java.util.List;
import java.util.Map.Entry;

import org.playonline.ui.basic.map.MapAnimation;
import org.playonline.ui.basic.map.MapGestureDetectorOverlay;
import org.playonline.ui.basic.map.MyOnGestureListener;
import org.playonline.ui.basic.map.PlayOnlineGeoPoint;
import org.playonline.ui.basic.map.PlayOnlineMapActivity;
import org.playonline.ui.basic.map.PlayOnlineGeoPoint.GoogleMapsUtil;

import com.google.android.maps.GeoPoint;
import com.google.android.maps.MapView;
import com.google.android.maps.Overlay;

public class OverlayManager {

	private final 	PlayOnlineMapActivity 	mapActivity;
	private final 	MapView					mapView;
	private final 	List<Overlay> 			mapOverlays;
	private 		CharacterOverlay 		characterOverlay;


	private OverlayManager(PlayOnlineMapActivity mapActivity){
		this.mapActivity 	= mapActivity;
		this.mapView		= mapActivity.getMapView();
		this.mapOverlays 	= mapView.getOverlays();
		// add of gesture detector to mapActivity (to enable long press, and other interactions)
		addOverlay(new MapGestureDetectorOverlay(new MyOnGestureListener(mapActivity)));
		// init current location overlay
		initCurrentLocationOverlay();
		((MyCharacterOverlay)characterOverlay).setAlpha(MapAnimation.FULLY_OPAQUE_APLHA);
		// insert elements into hidden elements list
		initHiddenElementsOverlayList();
	}

	public static OverlayManager getInstance(PlayOnlineMapActivity mapActivity){
		return new OverlayManager(mapActivity);
	}


	public CharacterOverlay getCharacterOverlay(){
		return characterOverlay;
	}



	public void addOverlay(Overlay overlay){
		// get a lock on the list of overlays
		synchronized (mapOverlays) {
			mapOverlays.add(overlay);
		}
	}


	/**************************************************************************************************************/
	/***************************************    HiddenElementOverlay management      ******************************/
	/**************************************************************************************************************/
	
	
	public void addHiddenElementOverlay(int hiddenElementHashCode){
		// retrieve hidden element and remove it from the hidden element list.
		HiddenElementOverlay overlay  = HiddenElementOverlay.getElementFromHashCode(hiddenElementHashCode);
		if(overlay != null){
			addOverlay(overlay);
			// update immediately the overlay without needing a screen touch from the user
			mapView.postInvalidate();
		}
	}


	public void removeHiddenElementOverlay(int hiddenElementHashCode){
		Overlay toRemove = null;
		synchronized (mapOverlays) {
			for (Overlay item : mapOverlays) {
				if(item.hashCode() == hiddenElementHashCode){
						toRemove = item;
						break;
				}				
			}
			if(toRemove != null){
				mapOverlays.size();
				mapOverlays.remove(toRemove);
				mapOverlays.size();
				// TODO bug here, object is properly removed from map overlays, but removed from screen with postInvalidate()...
				// update immediately the overlay without needing a screen touch from the user
				mapView.postInvalidate();
			}
		}
	}


	public void catchHiddenElement(String name, String description, int marker, int hiddenElementHashCode){
		mapActivity.getNewElementsAlertDialog().displayNewElementInInventory(name,description, marker);
		// remove the required HiddenElementOverlay designed by its hashCode
		removeHiddenElementOverlay(hiddenElementHashCode);
	}


	private void initHiddenElementsOverlayList(){
		HiddenElementOverlay overlay1 = new HiddenElementOverlay(mapActivity, GoogleMapsUtil.getGeoPointFromCoordinates(-21.054542,55.223828), 
				"Kit d'analyse","Permet de connaître la toxicité d'un élément.");
		HiddenElementOverlay overlay2 = new HiddenElementOverlay(mapActivity, GoogleMapsUtil.getGeoPointFromCoordinates(-21.388216,55.646217), 
				"Livre","Ouvrage ancien sur la pollinisation.");
		HiddenElementOverlay.addHiddentElementOverlay(overlay1);
		HiddenElementOverlay.addHiddentElementOverlay(overlay2);

	}

	
	/**************************************************************************************************************/
	/***************************************    CharacterOverlay management      ******************************/
	/**************************************************************************************************************/

	public void updateCharactersOverlays(){
		removeAllCharactersOverlays();
		computeMyCharacterLocation();
		computeOtherLocations();
		// check whether there are hidden element to unblock
		HiddenElementOverlay.unblockElement(mapActivity);
		// display immediately the overlay without needing a screen touch from the user
		mapView.postInvalidate();
	}


	private void removeAllCharactersOverlays(){
		List<Overlay> toRemove = new ArrayList<Overlay>();
		// get a lock on the map overlays list
		synchronized (mapOverlays) {
			for (Overlay overlay : mapOverlays) {
				if(overlay instanceof CharacterOverlay){
					toRemove.add(overlay);
				}
			}
			mapOverlays.removeAll(toRemove);
		}
	}


	private void computeMyCharacterLocation() {
		GeoPoint myCharacterLocation = mapActivity.getCharactersManager().getMyCharacterLocation();
		List<Short> charactersInMyLocation = mapActivity.getCharactersManager().getCharactersListInLocation(myCharacterLocation);
		// if my character is alone on its location
		if(charactersInMyLocation == null){
			characterOverlay = new MyCharacterOverlay(mapActivity, myCharacterLocation);
		}// otherwise
		else{
			characterOverlay = new MyCharacterOverlayShared(mapActivity, myCharacterLocation);
		}
		addOverlay(characterOverlay.getItemizedOverlay());
	}


	private void computeOtherLocations() {
		List<Short> charactersList;
		// for each location...
		for (Entry<GeoPoint,List<Short>> charactersListInLocation : mapActivity.getCharactersManager().getCharactersListInLocationMap().entrySet()) {
			// except the location of its own character which is already treated
			if(!charactersListInLocation.getKey().equals(mapActivity.getCharactersManager().getMyCharacterLocation())){
				charactersList = charactersListInLocation.getValue();
				if(charactersList.size() == 1){
					addOverlay(new OtherCharacterOverlay(mapActivity, charactersListInLocation.getKey()));
				}else if (charactersList.size() > 1){
					addOverlay(new OtherCharacterOverlayShared(mapActivity, charactersListInLocation.getKey()));
				}
			}
		}
	}

	
	private void initCurrentLocationOverlay(){
		// create an overlay for player location
		characterOverlay = new MyCharacterOverlay(mapActivity,PlayOnlineGeoPoint.BEGINNING_AREA);
		addOverlay(characterOverlay.getItemizedOverlay());
	}
}
